#  ============================================================================
#
#  Copyright (C) 2007-2008 Conceptive Engineering bvba. All rights reserved.
#  www.conceptive.be / project-camelot@conceptive.be
#
#  This file is part of the Camelot Library.
#
#  This file may be used under the terms of the GNU General Public
#  License version 2.0 as published by the Free Software Foundation
#  and appearing in the file LICENSE.GPL included in the packaging of
#  this file.  Please review the following information to ensure GNU
#  General Public Licensing requirements will be met:
#  http://www.trolltech.com/products/qt/opensource.html
#
#  If you are unsure which license is appropriate for your use, please
#  review the following information:
#  http://www.trolltech.com/products/qt/licensing.html or contact
#  project-camelot@conceptive.be.
#
#  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
#  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  For use of this library in commercial applications, please contact
#  project-camelot@conceptive.be
#
#  ============================================================================

import logging
logger = logging.getLogger('camelot.admin.validator.entity_validator')

from object_validator import ObjectValidator


class EntityValidator(ObjectValidator):
    """A validator class validates an entity before flushing it to the database
    and provides the user with feedback if the entity is not ready to flush
    """

    def objectValidity(self, entity_instance):
        """:return: list of messages explaining invalid data
        empty list if object is valid
        """
        from camelot.view.controls import delegates
        messages = []
        fields_and_attributes = dict(self.admin.get_columns())
        fields_and_attributes.update(dict(self.admin.get_fields()))
        for field, attributes in fields_and_attributes.items():
            # if the field was not editable, don't waste any time
            if attributes['editable']:
              value = getattr(entity_instance, field)
              #@todo: check if field is a primary key instead of checking 
              # whether the name is id
              if attributes['nullable']!=True and field!='id':
                  logger.debug('column %s is required'%(field))
                  if 'delegate' not in attributes:
                      raise Exception('no delegate specified for %s'%(field))
                  is_null = False
                  if value==None:
                      is_null = True
                  elif (attributes['delegate'] == delegates.CodeDelegate) and \
                       (sum(len(c) for c in value) == 0):
                      is_null = True
                  elif (attributes['delegate'] == delegates.PlainTextDelegate) and (len(value) == 0):
                      is_null = True
                  elif (attributes['delegate'] == delegates.Many2OneDelegate) and (not value.id):
                      is_null = True
                  elif (attributes['delegate'] == delegates.VirtualAddressDelegate) and (not value[1]):
                      is_null = True                    
                  if is_null:
                      messages.append(u'%s is a required field' % (attributes['name']))
        logger.debug(u'messages : %s'%(u','.join(messages)))
        return messages



#  ============================================================================
#
#  Copyright (C) 2007-2008 Conceptive Engineering bvba. All rights reserved.
#  www.conceptive.be / project-camelot@conceptive.be
#
#  This file is part of the Camelot Library.
#
#  This file may be used under the terms of the GNU General Public
#  License version 2.0 as published by the Free Software Foundation
#  and appearing in the file LICENSE.GPL included in the packaging of
#  this file.  Please review the following information to ensure GNU
#  General Public Licensing requirements will be met:
#  http://www.trolltech.com/products/qt/opensource.html
#
#  If you are unsure which license is appropriate for your use, please
#  review the following information:
#  http://www.trolltech.com/products/qt/licensing.html or contact
#  project-camelot@conceptive.be.
#
#  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
#  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  For use of this library in commercial applications, please contact
#  project-camelot@conceptive.be
#
#  ============================================================================

"""Helper functions for the view subpackage"""

from PyQt4 import QtCore

from datetime import datetime, time, date
import re

from camelot.core import constants

_local_date_format = None
 
def local_date_format():
    """Get the local data format and cache it for reuse"""
    global _local_date_format
    if not _local_date_format:
        locale = QtCore.QLocale()
        format_sequence = re.split('y*', str(locale.dateFormat(locale.ShortFormat)))
        # make sure a year always has 4 numbers
        format_sequence.insert(-1, 'yyyy')
        _local_date_format = ''.join(format_sequence)
    return _local_date_format

class ParsingError(Exception): pass

def string_from_string(s):
    if not s:
        return None
    return unicode(s)

def bool_from_string(s):
    if s is None: raise ParsingError()
    if s.lower() not in ['false', 'true']: raise ParsingError()
    return eval(s.lower().capitalize())


def date_from_string(s):
    from PyQt4.QtCore import QDate
    s = s.strip()
    if not s:
        return None
    try:
        dt = QDate.fromString(s, local_date_format())
    except ValueError:
        raise ParsingError()
    return date(dt.year(), dt.month(), dt.day())


def time_from_string(s, format=constants.strftime_time_format):
    if s is None: raise ParsingError()
    s = s.strip()

    try:
        dt = datetime.strptime(s, format)
    except ValueError:
        raise ParsingError()
    return dt.time()


def datetime_from_string(s, format=constants.strftime_datetime_format):
    if s is None: raise ParsingError()
    s = s.strip()

    try:
        dt = datetime.strptime(s, format)
    except ValueError:
        raise ParsingError()
    return dt


def int_from_string(s):
    if s is None: raise ParsingError()
    if s.isspace(): return int()

    s = s.strip()
    if len(s) == 0: return int()

    try:
        i = int(s)
    except ValueError:
        raise ParsingError()
    return i


def float_from_string(s):
    if not s:
        return None
    locale = QtCore.QLocale()
    f, ok = locale.toFloat(s)
    if not ok:
        raise ParsingError
    return f


def pyvalue_from_string(pytype, s):
    if pytype is str:
        return str(s)
    elif pytype is unicode:
        return unicode(s)
    elif pytype is bool:
        return bool_from_string(s)
    elif pytype is date:
        return date_from_string(s)
    elif pytype is time:
        return date_from_string(s)
    elif pytype is datetime:
        return datetime_from_string(s)
    elif pytype is float:
        return float_from_string(s)
    elif pytype is int:
        return int_from_string(s)



#  ============================================================================
#
#  Copyright (C) 2007-2008 Conceptive Engineering bvba. All rights reserved.
#  www.conceptive.be / project-camelot@conceptive.be
#
#  This file is part of the Camelot Library.
#
#  This file may be used under the terms of the GNU General Public
#  License version 2.0 as published by the Free Software Foundation
#  and appearing in the file LICENSE.GPL included in the packaging of
#  this file.  Please review the following information to ensure GNU
#  General Public Licensing requirements will be met:
#  http://www.trolltech.com/products/qt/opensource.html
#
#  If you are unsure which license is appropriate for your use, please
#  review the following information:
#  http://www.trolltech.com/products/qt/licensing.html or contact
#  project-camelot@conceptive.be.
#
#  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
#  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  For use of this library in commercial applications, please contact
#  project-camelot@conceptive.be
#
#  ============================================================================

from customdelegate import CustomDelegate, DocumentationMetaclass
from PyQt4 import QtGui, QtCore
from PyQt4.QtGui import QComboBox, QItemDelegate
from PyQt4.QtCore import QVariant, QString, Qt

from camelot.view.model_thread import post
from camelot.view.controls import editors
from camelot.core.utils import variant_to_pyobject
from camelot.view.proxy import ValueLoading

class ComboBoxDelegate(CustomDelegate):

    __metaclass__ = DocumentationMetaclass
    editor = editors.ChoicesEditor

    def __init__(self, parent, choices, editable=True, **kwargs):
        CustomDelegate.__init__(self, parent, editable=editable, **kwargs)
        self.choices = choices

    def setEditorData(self, editor, index):
        value = variant_to_pyobject(index.data(Qt.EditRole))
        editor.set_value(value)

        if callable(self.choices):
            def create_choices_getter(model, row):
                def choices_getter():
                    return list(self.choices(model._get_object(row)))
                return choices_getter
            post(create_choices_getter(index.model(), index.row()),
                 editor.set_choices)
        else:
            editor.set_choices(self.choices)

    def paint(self, painter, option, index):
        painter.save()
        self.drawBackground(painter, option, index)
        value = variant_to_pyobject(index.data(Qt.EditRole))
        if value in (None, ValueLoading):
            value = ''
        c = index.model().data(index, Qt.BackgroundRole)

        # let us be safe Qt.BackgroundRole valid only if set
        if c.type() == QVariant.Invalid:
            background_color = QtGui.QBrush()
        else:
            background_color = QtGui.QColor(c)

        rect = option.rect
        rect = QtCore.QRect(rect.left() + 3,
                            rect.top() + 6,
                            rect.width() - 5,
                            rect.height())

        if (option.state & QtGui.QStyle.State_Selected):
            painter.fillRect(option.rect, option.palette.highlight())
            fontColor = QtGui.QColor()
            if self.editable:
                Color = option.palette.highlightedText().color()
                fontColor.setRgb(Color.red(), Color.green(), Color.blue())
            else:
                fontColor.setRgb(130, 130, 130)
        else:
            if self.editable:
                painter.fillRect(option.rect, background_color)
                fontColor = QtGui.QColor()
                fontColor.setRgb(0, 0, 0)
            else:
                painter.fillRect(option.rect, option.palette.window())
                fontColor = QtGui.QColor()
                fontColor.setRgb(130, 130, 130)

        painter.setPen(fontColor.toRgb())
        rect = QtCore.QRect(option.rect.left()+2,
                            option.rect.top(),
                            option.rect.width()-4,
                            option.rect.height())
        painter.drawText(rect.x(),
                         rect.y(),
                         rect.width(),
                         rect.height(),
                         Qt.AlignVCenter | Qt.AlignLeft,
                         unicode(value))
        painter.restore()


class ComboBoxEditorDelegate(ComboBoxDelegate):
    """Delegate for combobox which makes sure that the editor (of the combobox)
is visible even if it isn't selected"""

    def __init__(self, choices, parent=None, editable=True, **kwargs):
        ComboBoxDelegate.__init__(self, parent, choices, editable, **kwargs)
        print 'choices by construction', choices(None)

    def setEditorData(self, editor, index):
        value = variant_to_pyobject(index.data(Qt.EditRole))

        def create_choices_getter(model, row):

            def choices_getter():
                print 'in set editor data', self.choices(None)
                return list(self.choices(None))

            return choices_getter

        #editor.set_value(value)
        post(create_choices_getter(index.model(),index.row()),
             editor.set_choices)

    def setModelData(self, editor, model, index):
        return None
        #model.setData(index, create_constant_function(editor.get_value()))

    def paint(self, painter, option, index):
        """adapted from booldelegate"""

        painter.save()
        self.drawBackground(painter, option, index)
        #checked = index.model().data(index, Qt.EditRole).toString().__str__()
        value = index.model().data(index, Qt.DisplayRole).toString().__str__()
        #c = index.model().data(index, Qt.BackgroundRole)
        #background_color = QtGui.QColor(c)
        background_color = QtGui.QColor(Qt.blue)

        #check_option = QtGui.QStyleOptionComboBox()
        #check_option = QtGui.QStyleOptionButton()
        check_option = QtGui.QStyleOptionMenuItem()
        #check_option.OptionType = QStyleOption.SO_ComboBox

        #check_option.text = QString('choice')

        rect = QtCore.QRect(option.rect.left(),
                            option.rect.top(),
                            option.rect.width(),
                            option.rect.height())

        #check_option.rect = rect
        check_option.palette = option.palette

        #if (option.state & QtGui.QStyle.State_Selected):
        #    painter.fillRect(option.rect, option.palette.highlight())
        #elif not self.editable:
        #    painter.fillRect(option.rect, option.palette.window())
        #else:
        #    painter.fillRect(option.rect, background_color)

        #if checked:
        #    check_option.state = option.state | QtGui.QStyle.State_On
        #else:
        #    check_option.state = option.state | QtGui.QStyle.State_Off
        check_option.state = QtGui.QStyle.State_DownArrow

        QtGui.QApplication.style().drawControl(QtGui.QStyle.CE_MenuBarItem,
                                               check_option,
                                               painter)

        QtGui.QApplication.style().drawItemText(painter,
                                                rect,
                                                0,
                                                option.palette,
                                                True,
                                                QString('choice'))

        painter.restore()


class TestComboBoxDelegate(QItemDelegate):

    def __init__(self, choices, parent=None):
        QItemDelegate.__init__(self, parent)
        self.choices = choices

    def createEditor(self, parent, option, index):
        editor = QComboBox(parent)
        i = 0
        for choice in self.choices:
            editor.insertItem(i, unicode(choice), QVariant(QString(choice)))
            i = i + 1
        return editor

    def setEditorData(self, comboBox, index):
        value = index.model().data(index, Qt.DisplayRole).toInt()
        comboBox.setCurrentIndex(value[0])

    def setModelData(self, editor, model, index):
        value = editor.currentIndex()
        model.setData(index, editor.itemData(value, Qt.DisplayRole))

    def updateEditorGeometry(self, editor, option, index):
        editor.setGeometry(option.rect)



#  ============================================================================
#
#  Copyright (C) 2007-2008 Conceptive Engineering bvba. All rights reserved.
#  www.conceptive.be / project-camelot@conceptive.be
#
#  This file is part of the Camelot Library.
#
#  This file may be used under the terms of the GNU General Public
#  License version 2.0 as published by the Free Software Foundation
#  and appearing in the file LICENSE.GPL included in the packaging of
#  this file.  Please review the following information to ensure GNU
#  General Public Licensing requirements will be met:
#  http://www.trolltech.com/products/qt/opensource.html
#
#  If you are unsure which license is appropriate for your use, please
#  review the following information:
#  http://www.trolltech.com/products/qt/licensing.html or contact
#  project-camelot@conceptive.be.
#
#  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
#  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  For use of this library in commercial applications, please contact
#  project-camelot@conceptive.be
#
#  ============================================================================

"""
Python structures to represent filters.
These structures can be transformed to QT forms.
"""

from camelot.view.model_thread import gui_function
from camelot.core.utils import ugettext_lazy as _

def structure_to_filter(structure):
    """Convert a python data structure to a filter, using the following rules :
    
    if structure is an instance of Filter, return structure
    else create a GroupBoxFilter from the structure
    """
    if isinstance(structure, Filter):
        return structure
    return GroupBoxFilter(structure)
  
class Filter(object):
    """Base class for filters"""
    
    def __init__(self, attribute, value_to_string=lambda x:unicode(x)):
        """
        @param attribute: the attribute on which to filter, this attribute
        may contain dots to indicate relationships that need to be followed, 
        eg.  'person.groups.name'
        @param value_to_string: function that converts a value of the attribute to
        a string that will be displayed in the filter 
        """
        self.attribute = attribute
        self._value_to_string = value_to_string
    
    @gui_function     
    def render(self, parent, name, options):
        """Render this filter as a qt object
        @param parent: its parent widget
        @param name: the name of the filter
        @param options: the options that can be selected, where each option is a list
        of tuples containting (option_name, query_decorator)  
        
        The name and the list of options can be fetched with get_name_and_options"""
        raise NotImplementedError()
        
    def get_name_and_options(self, admin):
        """return a tuple of the name of the filter and a list of options that can be selected. 
        Each option is a tuple of the name of the option, and a filter function to
        decorate a query
        @return:  (filter_name, [(option_name, query_decorator), ...)
        """
        from sqlalchemy.sql import select
        from sqlalchemy import orm
        from elixir import session
        filter_names = []
        joins = []
        table = admin.entity.table
        path = self.attribute.split('.')
        for field_name in path:
            attributes = admin.get_field_attributes(field_name)
            filter_names.append(attributes['name'])
            # @todo: if the filter is not on an attribute of the relation, but on the relation itselves
            if 'target' in attributes:
                admin = attributes['admin']
                joins.append(field_name)
                if attributes['direction'] == orm.interfaces.MANYTOONE:
                    table = admin.entity.table.join(table)
                else:
                    table = admin.entity.table
                  
          
        col = getattr(admin.entity, field_name)
        query = select([col], distinct=True, order_by=col.asc()).select_from(table)
          
        def create_decorator(col, value, joins):
          
            def decorator(q):
                if joins:
                    q = q.join(joins, aliased=True)
                return q.filter(col==value)
              
            return decorator
      
        options = [(self._value_to_string(value[0]), create_decorator(col, value[0], joins))
                   for value in session.execute(query)]
    
        return (filter_names[0],[(_('all'), lambda q: q)] + options)
        
class GroupBoxFilter(Filter):
    """Filter where the items are displayed in a QGroupBox"""
    
    @gui_function
    def render(self, parent, name, options):
      
        from PyQt4 import QtCore, QtGui
        from camelot.view.controls.filterlist import filter_changed_signal
        
        class FilterWidget(QtGui.QGroupBox):
            """A box containing a filter that can be applied on a table view, this filter is
            based on the distinct values in a certain column"""
          
            def __init__(self, name, choices, parent):
                QtGui.QGroupBox.__init__(self, unicode(name), parent)
                self.group = QtGui.QButtonGroup(self)
                self.item = name
                self.unique_values = []
                self.choices = None
                self.setChoices(choices)
                 
            def emit_filter_changed(self, state):
                self.emit(filter_changed_signal)
            
            def setChoices(self, choices):
                self.choices = choices
                layout = QtGui.QVBoxLayout()
                for i,name in enumerate([unicode(c[0]) for c in choices]):
                    button = QtGui.QRadioButton(name, self)
                    layout.addWidget(button)
                    self.group.addButton(button, i)
                    if i==0:
                        button.setChecked(True)
                    self.connect(button, QtCore.SIGNAL('toggled(bool)'), self.emit_filter_changed)
                layout.addStretch()
                self.setLayout(layout)
            
            def decorate_query(self, query):
                checked = self.group.checkedId()
                if checked>=0:
                    return self.choices[checked][1](query)
                return query
                
        return FilterWidget(name, options, parent)
      
class ComboBoxFilter(Filter):
    """Filter where the items are displayed in a QComboBox"""
    
    @gui_function
    def render(self, parent, name, options):
      
        from PyQt4 import QtCore, QtGui
        from camelot.view.controls.filterlist import filter_changed_signal
        
        class FilterWidget(QtGui.QGroupBox):
          
            def __init__(self, name, choices, parent):
                QtGui.QGroupBox.__init__(self, name, parent)
                layout = QtGui.QVBoxLayout()
                self.choices = choices
                combobox = QtGui.QComboBox(self)
                for i,(name,decorator) in enumerate(choices):
                    combobox.insertItem(i, unicode(name), QtCore.QVariant(decorator))
                layout.addWidget(combobox)
                self.setLayout(layout)
                self.current_index = 0
                self.connect(combobox, QtCore.SIGNAL('currentIndexChanged(int)'), self.emit_filter_changed)
                    
            def emit_filter_changed(self, index):
                self.current_index = index
                self.emit(filter_changed_signal)
                
            def decorate_query(self, query):
                if self.current_index>=0:
                    return self.choices[self.current_index][1](query)
                return query
              
        return FilterWidget(name, options, parent)
    
class ValidDateFilter(Filter):
    """Filters entities that are valid a certain date.  This filter will present
    a date to the user and filter the entities that have their from date before this
    date and their end date after this date.  If no date is given, all entities will
    be shown"""

    def __init__(self, from_attribute, thru_attribute, verbose_name):
        """
        :param from_attribute: the name of the attribute representing the from date
        :param thru_attribute: the name of the attribute representing the thru date
        :param verbose_name: the displayed name of the filter"""
        self._from_attribute = from_attribute
        self._thru_attribute = thru_attribute
        self._verbose_name = verbose_name
        
    def render(self, parent, name, options):
        
        from datetime import date
        from PyQt4 import QtGui, QtCore
        from camelot.view.controls.filterlist import filter_changed_signal
        from camelot.view.controls.editors import DateEditor, editingFinished
        
        class FilterWidget(QtGui.QGroupBox):
          
            def __init__(self, name, query_decorator, parent):
                QtGui.QGroupBox.__init__(self, name, parent)
                layout = QtGui.QVBoxLayout()
                self.date_editor = DateEditor(parent=self, nullable=True)
                self.date_editor.set_value(date.today())
                self.query_decorator = query_decorator
                layout.addWidget(self.date_editor)
                self.setLayout(layout)
                self.connect(self.date_editor, editingFinished, self.emit_filter_changed)
                    
            def emit_filter_changed(self):
                self.emit(filter_changed_signal)
                
            def decorate_query(self, query):
                return self.query_decorator(query, self.date_editor.get_value())
              
        return FilterWidget(name, options, parent)
        
    def get_name_and_options(self, admin):
        from sqlalchemy.sql import and_
        
        def query_decorator(query, date):
            e = admin.entity
            if date:
                print 'filter on', date
                return query.filter(and_(getattr(e, self._from_attribute)<=date,
                                         getattr(e, self._thru_attribute)>=date))
            return query
        
        return (self._verbose_name, query_decorator)


#  ============================================================================
#
#  Copyright (C) 2007-2008 Conceptive Engineering bvba. All rights reserved.
#  www.conceptive.be / project-camelot@conceptive.be
#
#  This file is part of the Camelot Library.
#
#  This file may be used under the terms of the GNU General Public
#  License version 2.0 as published by the Free Software Foundation
#  and appearing in the file LICENSE.GPL included in the packaging of
#  this file.  Please review the following information to ensure GNU
#  General Public Licensing requirements will be met:
#  http://www.trolltech.com/products/qt/opensource.html
#
#  If you are unsure which license is appropriate for your use, please
#  review the following information:
#  http://www.trolltech.com/products/qt/licensing.html or contact
#  project-camelot@conceptive.be.
#
#  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
#  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  For use of this library in commercial applications, please contact
#  project-camelot@conceptive.be
#
#  ============================================================================

"""Manage QAction objects"""

from PyQt4 import QtGui, QtCore

def getAction(parent, widgetaction):
    if widgetaction:
        return QtGui.QWidgetAction(parent)
    else:
        return QtGui.QAction(parent)

def createAction(*a, **kw):
    """creates and returns a QAction object"""

    # collect params
    parent = kw['parent']
    text = kw['text']
    slot = kw.get('slot', None)
    shortcut = kw.get('shortcut', '')
    actionicon = kw.get('actionicon', '')
    tip = kw.get('tip', '')
    checkable = kw.get('checkable', False)
    signal = kw.get('signal', 'triggered()')
    widgetaction = kw.get('widgetaction', False)

    action = getAction(parent, widgetaction)

    action.setText(text)

    if actionicon:
        action.setIcon(QtGui.QIcon(actionicon))
    if shortcut:
        action.setShortcut(shortcut)
    if tip:
        action.setToolTip(tip)
        action.setStatusTip(tip)
    if slot is not None:
        parent.connect(action, QtCore.SIGNAL(signal), slot)
    if checkable:
        action.setCheckable(True)
    return action

def addActions(target, actions):
    """add action objects to menus, menubars, and toolbars
    if action is None, add a separator.
    """
    for action in actions:
        if action is None:
            target.addSeparator()
        else:
            target.addAction(action)



#import urllib
#from PyQt4 import QtGui, QtCore
#
#class NewCompanyWizard( QtGui.QWizard ):
#    def __init__( wizard, organization_getter, parent = None ):
#        super( NewCompanyWizard, wizard ).__init__( parent )
#        wizard.value = ''
#        wizard.success = False
#        wizard.name = ''
#        wizard.town = ''
#        wizard.street = ''
#        wizard.taxid = ''
#        wizard.setWindowTitle( 'New Company Wizard' )
#        combobox = QtGui.QComboBox()
#        combobox.insertItem( 0, unicode( 'Loading list ...' ), QtCore.QVariant( all ) )
#
#        class WelcomePage( QtGui.QWizardPage ):
#            def __init__( self ):
#                super( WelcomePage, self ).__init__()
#                self.setTitle( 'New Company Wizard' )
#                self.setSubTitle( 'Welcome, This wizard will help you to add a new company' )
#
#        class inputPage( QtGui.QWizardPage ):
#            def __init__( self ):
#                super( inputPage, self ).__init__()
#                self.comboBox = QtGui.QComboBox()
#                self.vat = QtGui.QLineEdit()
#                self.setTitle( 'New Company Wizard' )
#                self.setSubTitle( 'Select the company\'s country, and fill in its VAT Number' )
#                layout = QtGui.QHBoxLayout()
#                layout.addSpacing( 30 )
#                layout.addWidget( self.comboBox )
#                layout.addWidget( self.vat )
#                #layout.maximumSize()
#                self.connect( self.vat, QtCore.SIGNAL( 'textChanged(const QString&)' ), self.textChanged )
#                layout.addStretch( 0 )
#                self.setLayout( layout )
#
#            def textChanged( self, text ):
#                self.emit( QtCore.SIGNAL( 'completeChanged()' ) )
#
#            def isComplete( self ):
#                wizard.value = str( self.comboBox.currentText() )[0:2] + str( self.vat.text() )
#                if wizard.value[10:] is '':
#                    return False
#                else:
#                    return True
#
#            def initializePage( self ):
#                landen = QtCore.QStringList()
#                landen.append( 'AT - Austria' )
#                landen.append( 'BE - Belgium' )
#                landen.append( 'BG - Bulgaria' )
#                landen.append( 'CY - Cyprus' )
#                landen.append( 'CZ - Czech Republic' )
#                landen.append( 'DE - Germany' )
#                landen.append( 'DK - Denmark' )
#                landen.append( 'EE - Estonia' )
#                landen.append( 'EL - Greece ' )
#                landen.append( 'ES - Spain ' )
#                landen.append( 'FI - Finland' )
#                landen.append( 'FR - France' )
#                landen.append( 'GB - United Kingdom' )
#                landen.append( 'HU - Hungary' )
#                landen.append( 'IE - Ireland' )
#                landen.append( 'IT - Italy' )
#                landen.append( 'LT - Lithuania' )
#                landen.append( 'LU - Luxembourg' )
#                landen.append( 'LV - Latvia' )
#                landen.append( 'MT - Malta' )
#                landen.append( 'NL - The Netherlands' )
#                landen.append( 'PL - Poland' )
#                landen.append( 'PT - Portugal' )
#                landen.append( 'RO - Romania' )
#                landen.append( 'SE - Sweden' )
#                landen.append( 'SI - Slovenia' )
#                landen.append( 'sK - Slovakia' )
#                self.comboBox.addItems( landen )
#
#        class checkPage( QtGui.QWizardPage ):
#            def __init__( self ):
#                super( checkPage, self ).__init__()
#                self.success = False
#                self.ondernemingsVorm = QtGui.QLabel()
#                self.name = QtGui.QLabel()
#                self.street = QtGui.QLabel()
#                self.town = QtGui.QLabel()
#                self.setTitle( 'Add New Company' )
#                if self.success == False:
#                    self.setSubTitle( '' )
#                else:
#                    self.setSubTitle( 'Here is the adress we found in the database, make sure this is correct.' )
#                layout = QtGui.QVBoxLayout()
#                naamlayout = QtGui.QHBoxLayout()
#                naamlayout.addWidget( self.ondernemingsVorm )
#                naamlayout.addWidget( self.name )
#                layout.addLayout( naamlayout )
#                layout.addWidget( self.street )
#                layout.addWidget( self.town )
#                layout.addStretch( 1 )
#                self.setLayout( layout )
#
#            def isComplete( self ):
#                if not wizard.success:
#                    return False
#                else:
#                    return True
#
#            def initializePage( self ):
#                url = 'http://ec.europa.eu/taxation_customs/vies/viesquer.do?ms=' + wizard.value[0:2] + '&iso=' + wizard.value[0:2] + '&vat=' + wizard.value[2:]
#                file = urllib.urlopen( url )
#                test = file.readlines()
#                succes = str( test[323] )
#                name = str( test[365] )
#                adress = str( test[378] )
#                succes = succes.strip()
#                print succes
#                if succes[0] == 'S':
#                    self.setSubTitle( 'System Overloaded, please try again later.' )
#                else:
#                    if succes[0] == 'Y':
#                        wizard.success = True
#                    else:
#                        wizard.success = False
#                    if wizard.success == False:
#                        self.setSubTitle( '' )
#                    else:
#                        self.setSubTitle( 'Here is the address we found in the database, make sure this is correct.' )
#                        name = name.strip()
#                        if name.find( ' ' ) == -1:
#                            self.setSubTitle( 'VAT Number Correct, but no information was found in the Database.' )
#                            self.name.setText( '' )
#                            self.ondernemingsVorm.setText( '' )
#                            self.street.setText( '' )
#                            self.town.setText( '' )
#                        else:
#                            self.name.setText( name )
#                            firstSpace = name.index( ' ' )
#                            self.ondernemingsVorm.setText( name[0:firstSpace] )
#                            name = name[firstSpace:]
#                            wizard.name = name.strip()
#                            self.name.setText( name )
#                            adress = adress.strip()
#                            firstTag = adress.index( '<' )
#                            wizard.street = adress[0:firstTag]
#                            adress = adress[firstTag + 4:]
#                            wizard.town = adress
#                            self.street.setText( wizard.street )
#                            self.town.setText( wizard.town )
#                            wizard.taxid = str( wizard.value )
#                    if not wizard.success:
#                        self.ondernemingsVorm.setText( 'Nothing Found.' )
#                        self.name.setText( ' ' )
#                        self.street.setText( 'Please retyp the VAT number' )
#                        self.town.setText( ' ' )
#
#        wizard.addPage( WelcomePage() )
#        wizard.addPage( inputPage() )
#        wizard.addPage( checkPage() )
#
#    def finished( self, result ):
#        pass



#  ============================================================================
#
#  Copyright (C) 2007-2008 Conceptive Engineering bvba. All rights reserved.
#  www.conceptive.be / project-camelot@conceptive.be
#
#  This file is part of the Camelot Library.
#
#  This file may be used under the terms of the GNU General Public
#  License version 2.0 as published by the Free Software Foundation
#  and appearing in the file LICENSE.GPL included in the packaging of
#  this file.  Please review the following information to ensure GNU
#  General Public Licensing requirements will be met:
#  http://www.trolltech.com/products/qt/opensource.html
#
#  If you are unsure which license is appropriate for your use, please
#  review the following information:
#  http://www.trolltech.com/products/qt/licensing.html or contact
#  project-camelot@conceptive.be.
#
#  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
#  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  For use of this library in commercial applications, please contact
#  project-camelot@conceptive.be
#
#  ============================================================================

"""Admin class for Plain Old Python Object"""

import logging
logger = logging.getLogger('camelot.view.object_admin')

from camelot.view.model_thread import gui_function, model_function
from camelot.core.utils import ugettext as _
from validator.object_validator import ObjectValidator


class ObjectAdmin(object):
    """The ObjectAdmin class describes the interface that will be used
    to interact with objects of a certain class.  The behaviour of this class
    and the resulting interface can be tuned by specifying specific class
    attributes:

    .. attribute:: verbose_name

    A human-readable name for the object, singular ::

    verbose_name = 'movie'

    If this isn't given, the class name will be used

    .. attribute:: verbose_name_plural

    A human-readable name for the object, plural ::

    verbose_name_plural = 'movies'

    If this isn't given, Camelot will use verbose_name + "s"

    .. attribute:: list_display

    a list with the fields that should be displayed in a table view

    .. attribute:: form_display

    a list with the fields that should be displayed in a form view, defaults to
    the same fields as those specified in list_display ::

    class Admin(EntityAdmin):
      form_display = ['title', 'rating', 'cover']

    instead of telling which forms to display. It is also possible to define
    the form itself ::

    from camelot.view.forms import Form, TabForm, WidgetOnlyForm, HBoxForm

    class Admin(EntityAdmin):
      form_display = TabForm([
        ('Movie', Form([
          HBoxForm([['title', 'rating'], WidgetOnlyForm('cover')]),
          'short_description',
          'releasedate',
          'director',
          'script',
          'genre',
          'description', 'tags'], scrollbars=True)),
        ('Cast', WidgetOnlyForm('cast'))
      ])


    .. attribute:: list_filter

    A list of fields that should be used to generate filters for in the table
    view.  If the field named is a one2many, many2one or many2many field, the
    field name should be followed by a field name of the related entity ::

    class Project(Entity):
      oranization = OneToMany('Organization')
      name = Field(Unicode(50))

      class Admin(EntityAdmin):
        list_display = ['organization']
        list_filter = ['organization.name']

    .. image:: ../_static/filter/group_box_filter.png

    .. attribute:: list_search

    A list of fields that should be searched when the user enters something in
    the search box in the table view.  By default only character fields are
    searched.  For use with one2many, many2one or many2many fields, the same
    rules as for the list_filter attribute apply

    .. attribute:: form_size

    a tuple indicating the size of a form view, defaults to (700,500)

    .. attribute:: form_actions

    Actions to be accessible by pushbuttons on the side of a form,
    a list of tuples (button_label, action_function) where action_function
    takes as its single argument, a method that returns the the object that
    was displayed by the form when the button was pressed::

    class Admin(EntityAdmin):
      form_actions = [('Foo', lamda o_getter:print 'foo')]

    .. attribute:: field_attributes

    A dictionary specifying for each field of the model some additional
    attributes on how they should be displayed.  All of these attributes
    are propagated to the constructor of the delegate of this field::

    class Movie(Entity):
      title = Field(Unicode(50))

      class Admin(EntityAdmin):
        list_display = ['title']
        field_attributes = dict(title=dict(editable=False))

    Other field attributes process by the admin interface are:

    .. attribute:: name
    The name of the field used, this defaults to the name of the attribute

    .. attribute:: target
    In case of relation fields, specifies the class that is at the other
    end of the relation.  Defaults to the one found by introspection.

    .. attribute:: admin
    In case of relation fields, specifies the admin class that is to be used
    to visualize the other end of the relation.  Defaults to the default admin
    class of the target class.
    """
    name = None #DEPRECATED
    verbose_name = None
    verbose_name_plural = None
    list_display = []
    validator = ObjectValidator
    fields = []
    form = [] #DEPRECATED
    form_display = []
    list_filter = []
    list_charts = []
    list_actions = []
    list_search = []
    list_size = (600, 400)
    form_size = (700, 500)
    form_actions = []
    form_title_column = None #DEPRECATED
    field_attributes = {}

    def __init__(self, app_admin, entity):
        """:param app_admin: the application admin object for this application
        
        :param entity: the entity class for which this admin instance is to be
        used
        """
        from camelot.view.remote_signals import get_signal_handler
        self.app_admin = app_admin
        self.rsh = get_signal_handler()
        if entity:
            from camelot.view.model_thread import get_model_thread
            self.entity = entity
            self.mt = get_model_thread()
        #
        # caches to prevent recalculation of things
        #
        self._field_attributes = dict()
        self._subclasses = None

    def __str__(self):
        return 'Admin %s' % str(self.entity.__name__)

    def __repr__(self):
        return 'ObjectAdmin(%s)' % str(self.entity.__name__)

    def get_name(self):
        return self.get_verbose_name()

    def get_verbose_name(self):
        return unicode(
            self.verbose_name or self.name or _(self.entity.__name__)
        )

    def get_verbose_name_plural(self):
        return unicode(
            self.verbose_name_plural
            or self.name 
            or (self.get_verbose_name() + 's')
        )

    @model_function
    def get_verbose_identifier(self, obj):
        """Create an identifier for an object that is interpretable
        for the user, eg : the 'id' of an object.  This verbose identifier can
        be used to generate a title for a form view of an object.
        """
        return u'%s : %s' % (self.get_verbose_name(), unicode(obj))

    def get_entity_admin(self, entity):
        return self.app_admin.get_entity_admin(entity)

    @model_function
    def get_form_actions(self, entity):
        from camelot.admin.form_action import structure_to_form_actions
        return structure_to_form_actions(self.form_actions)

    @model_function
    def get_list_actions(self):
        from camelot.admin.list_action import structure_to_list_actions
        return structure_to_list_actions(self.list_actions)

    @model_function
    def get_subclass_tree( self ):
        """Get a tree of admin classes representing the subclasses of the class
        represented by this admin class
        
        :return: [(subclass_admin, [(subsubclass_admin, [...]),...]),...]
        """
        subclasses = []
        for subclass in self.entity.__subclasses__():
            subclass_admin = self.get_related_entity_admin(subclass)
            subclasses.append((
                subclass_admin,
                subclass_admin.get_subclass_tree()
            ))
        return subclasses

    def get_related_entity_admin(self, entity):
        """Get an admin object for another entity.  Taking into account
        preferences of this admin object or for those of admin object higher up
        the chain such as the application admin object.

        :param entity: the entity class for which an admin object is requested
        """
        related_admin = self.app_admin.get_entity_admin(entity)
        if not related_admin:
            logger.warn('no related admin found for %s' % (entity.__name__))
        return related_admin

    def get_field_attributes(self, field_name):
        """Get the attributes needed to visualize the field field_name
        
        :param field_name : the name of the field
        
        :return: a dictionary of attributes needed to visualize the field,
        those attributes can be:
         * python_type : the corresponding python type of the object
         * editable : bool specifying wether the user can edit this field
         * widget : which widget to be used to render the field
         * ...
        """
        try:
            return self._field_attributes[field_name]
        except KeyError:
            from camelot.view.controls import delegates
            #
            # Default attributes for all fields
            #
            attributes = dict(
                python_type=str,
                length=None,
                tooltip=None,
                background_color=None,
                minimal_column_width=0,
                editable=False,
                nullable=True,
                widget='str',
                blank=True,
                delegate=delegates.PlainTextDelegate,
                validator_list=[],
                name=unicode(_(field_name.replace( '_', ' ' ))).capitalize()
            )
            #
            # Field attributes forced by the field_attributes property
            #
            forced_attributes = {}
            try:
                forced_attributes = self.field_attributes[field_name]
            except KeyError:
                pass

            #
            # TODO : move part of logic from entity admin class over here
            #

            #
            # Overrule introspected field_attributes with those defined
            #
            attributes.update(forced_attributes)

            #
            # In case of a 'target' field attribute, instantiate an appropriate
            # 'admin' attribute
            #

            def get_entity_admin(target):
                """Helper function that instantiated an Admin object for a
                target entity class

                :param target: an entity class for which an Admin object is
                needed
                """
                try:
                    fa = self.field_attributes[field_name]
                    target = fa.get('target', target)
                    admin_class = fa['admin']
                    return admin_class(self.app_admin, target)
                except KeyError:
                    return self.get_related_entity_admin(target)

            if 'target' in attributes:
                attributes['admin'] = get_entity_admin(attributes['target'])

            # if name should be translated, do so now
            attributes['name'] = unicode(attributes['name'])
            self._field_attributes[field_name] = attributes
            return attributes

    @model_function
    def get_columns(self):
        """
        The columns to be displayed in the list view, returns a list of pairs
        of the name of the field and its attributes needed to display it
        properly

        @return: [(field_name,
                  {'widget': widget_type,
                   'editable': True or False,
                   'blank': True or False,
                   'validator_list':[...],
                   'name':'Field name'}),
                 ...]
        """
        return [(field, self.get_field_attributes(field))
                for field in self.list_display]

    def create_validator(self, model):
        return self.validator(self, model)

    @model_function
    def get_fields(self):
        if self.form or self.form_display:
            fields = self.get_form_display().get_fields()
        elif self.fields:
            fields = self.fields
        else:
            fields = self.list_display
        fields_and_attributes =  [
                (field, self.get_field_attributes(field))
                for field in fields
        ]
        return fields_and_attributes

    @model_function
    def get_all_fields_and_attributes(self):
        """A list of (field_name, field_attributes) for all fields that can
        possibly appear in a list or a form or for which field attributes have
        been defined
        """
        pass

    @model_function
    def get_form_display(self):
        from camelot.view.forms import Form, structure_to_form
        if self.form or self.form_display:
            return structure_to_form(self.form or self.form_display)
        if self.list_display:
            return Form(self.list_display)
        return Form([])

    @gui_function
    def create_form_view(self, title, model, index, parent):
        """Creates a Qt widget containing a form view, for a specific index in
        a model; uses the Admin class"""
        logger.debug('creating form view for index %s' % index)
        from camelot.view.controls.formview import FormView
        form = FormView(title, self, model, index)
        return form
    
    @model_function
    def delete(self, entity_instance):
        """Delete an entity instance"""
        del entity_instance
        
    @model_function
    def flush(self, entity_instance):
        """Flush the pending changes of this entity instance to the backend"""
        pass
    
    @model_function
    def add(self, entity_instance):
        """Add an entity instance as a managed entity instance"""
        pass



from PyQt4 import QtGui, QtCore
from PyQt4.QtCore import Qt

from customdelegate import CustomDelegate
from camelot.view.controls import editors
from camelot.core.utils import variant_to_pyobject, create_constant_function
from camelot.view.proxy import ValueLoading

import logging
logger = logging.getLogger('camelot.view.controls.delegates.many2onedelegate')

class Many2OneDelegate(CustomDelegate):
    """Custom delegate for many 2 one relations
    
  .. image:: ../_static/manytoone.png
  """
  
    editor = editors.Many2OneEditor
                                    
    def __init__(self,
                 parent=None,
                 admin=None,
                 embedded=False,
                 editable=True,
                 **kwargs):
        logger.debug('create many2onecolumn delegate')
        assert admin != None
        CustomDelegate.__init__(self, parent, editable, **kwargs)
        self.admin = admin
        self._embedded = embedded
        self._kwargs = kwargs
        self._width = self._width * 2    
    
    def paint(self, painter, option, index):
        painter.save()
        self.drawBackground(painter, option, index)
        
        
        
        value = index.data(Qt.DisplayRole).toString()
        
        background_color = QtGui.QColor(index.model().data(index, Qt.BackgroundRole))
        
        if( option.state & QtGui.QStyle.State_Selected ):
            painter.fillRect(option.rect, option.palette.highlight())
            fontColor = QtGui.QColor()
            if self.editable:
                Color = option.palette.highlightedText().color()
                fontColor.setRgb(Color.red(), Color.green(), Color.blue())
            else:
                fontColor.setRgb(130,130,130)
        else:
            if self.editable:
                painter.fillRect(option.rect, background_color)
                fontColor = QtGui.QColor()
                fontColor.setRgb(0,0,0)
            else:
                painter.fillRect(option.rect, option.palette.window())
                fontColor = QtGui.QColor()
                fontColor.setRgb(130,130,130)
                
        painter.setPen(fontColor.toRgb())
        
        painter.drawText(option.rect.x()+2,
                         option.rect.y(),
                         option.rect.width()-4,
                         option.rect.height(),
                         Qt.AlignVCenter | Qt.AlignLeft,
                         unicode(value))
        
        
        painter.restore()
        
    def createEditor(self, parent, option, index):
        if self._embedded:
            editor = editors.EmbeddedMany2OneEditor(self.admin, parent, editable=self.editable)
        else:
            editor = editors.Many2OneEditor(self.admin, parent, editable=self.editable)
        if option.version != 5:
            editor.setAutoFillBackground(True)            
        self.connect(editor,
                     QtCore.SIGNAL('editingFinished()'),
                     self.commitAndCloseEditor)
        return editor
    
    def setEditorData(self, editor, index):
        value = variant_to_pyobject(index.data(Qt.EditRole))
        if value!=ValueLoading:
            editor.set_value(create_constant_function(value))
        else:
            editor.set_value(ValueLoading)
      
    def setModelData(self, editor, model, index):
        if editor.entity_instance_getter:
            model.setData(index, editor.entity_instance_getter)
          
#  def sizeHint(self, option, index):
#    return self._dummy_editor.sizeHint()    



#  ============================================================================
#
#  Copyright (C) 2007-2008 Conceptive Engineering bvba. All rights reserved.
#  www.conceptive.be / project-camelot@conceptive.be
#
#  This file is part of the Camelot Library.
#
#  This file may be used under the terms of the GNU General Public
#  License version 2.0 as published by the Free Software Foundation
#  and appearing in the file LICENSE.GPL included in the packaging of
#  this file.  Please review the following information to ensure GNU
#  General Public Licensing requirements will be met:
#  http://www.trolltech.com/products/qt/opensource.html
#
#  If you are unsure which license is appropriate for your use, please
#  review the following information:
#  http://www.trolltech.com/products/qt/licensing.html or contact
#  project-camelot@conceptive.be.
#
#  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
#  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  For use of this library in commercial applications, please contact
#  project-camelot@conceptive.be
#
#  ============================================================================

"""
Tool to assist in the creation and management of Camelot
projects.

use : python camelot-admin.py -h 

to get a list of commands and options
"""

def startproject(project):
    import shutil, os, sys
    if os.path.exists(project):
        raise Exception('Directory %s allready exists, cannot start a project in it'%project)
    
    def ignore(directory, content):
        """ignore .svn files"""
        for c in content:
            if c.startswith('.'):
                yield c
                
    # ignore is only supported as of python 2.6
    v = sys.version_info
    if v[0]>2 or (v[0]==2 and v[1]>=6):
        print 'ignore .svn files'
        shutil.copytree(os.path.join(os.path.dirname(__file__), '..', 'empty_project'), 
                        project, ignore=ignore)
    else:
        shutil.copytree(os.path.join(os.path.dirname(__file__), '..', 'empty_project'), 
                        project)    
    # creating a repository doesn't seems to work when migrate is easy intalled
    #from migrate.versioning.api import create
    #create(os.path.join(project, 'repository'), project)
      
commands = locals()

def main():
    from optparse import OptionParser
    parser = OptionParser(usage='usage: %prog [options] startproject project')
    (options, args) = parser.parse_args()
    command, project = args
    commands[command](*args[1:])
    
if __name__ == '__main__':
    main()



import settings
import logging

logger = logging.getLogger('main')

if __name__ == '__main__':
    from camelot.view.main import main
    from application_admin import MyApplicationAdmin
    main(MyApplicationAdmin())



#  ============================================================================
#
#  Copyright (C) 2007-2008 Conceptive Engineering bvba. All rights reserved.
#  www.conceptive.be / project-camelot@conceptive.be
#
#  This file is part of the Camelot Library.
#
#  This file may be used under the terms of the GNU General Public
#  License version 2.0 as published by the Free Software Foundation
#  and appearing in the file LICENSE.GPL included in the packaging of
#  this file.  Please review the following information to ensure GNU
#  General Public Licensing requirements will be met:
#  http://www.trolltech.com/products/qt/opensource.html
#
#  If you are unsure which license is appropriate for your use, please
#  review the following information:
#  http://www.trolltech.com/products/qt/licensing.html or contact
#  project-camelot@conceptive.be.
#
#  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
#  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  For use of this library in commercial applications, please contact
#  project-camelot@conceptive.be
#
#  ============================================================================


class fifo(object):
    """fifo, is the actual cache containing a limited set of copies of row data
    so the data in fifo, is always immediately accessible to the gui thread,
    with zero delay as you scroll down the table view, fifo is filled and
    refilled with data queried from the database
    
    the cache can be queried either by the row number or by the primary key
    of the object represented by the row data.
    """
    def __init__(self, max_entries):
        self.max_entries = max_entries
        self.entities = []
        self.data_by_rows = dict()
        self.rows_by_entity = dict()
        
    def __unicode__(self):
        return u','.join(unicode(e) for e in self.entities)
    
    def add_data(self, row, entity, value):
        self.delete_by_entity(entity)
        self.data_by_rows[row] = (entity, value)
        self.rows_by_entity[entity] = row
        self.entities.append(entity)
        if len(self.entities)>self.max_entries:
            entity = self.entities.pop(0)
            row = self.rows_by_entity[entity]
            del self.data_by_rows[row]
            del self.rows_by_entity[entity]
      
    def delete_by_row(self, row):
        (entity, value) = self.data_by_rows[row]
        del self.data_by_rows[row]
        del self.rows_by_entity[entity] 
        return row
    
    def delete_by_entity(self, entity):
        """Remove everything in the cache related to an entity instance
        returns the row at which the data was stored if the data was in the
        cache, return None otherwise"""
        row = None
        try:
            row = self.rows_by_entity[entity]
            del self.data_by_rows[row]
            del self.rows_by_entity[entity]      
        except KeyError:
            pass
        try:
            self.entities.remove(entity)
        except ValueError:
            pass
        return row    
    
    def get_data_at_row(self, row):
        return self.data_by_rows[row][1]
    
    def get_row_by_entity(self, entity):
        return self.rows_by_entity[entity]
    
    def get_entity_at_row(self, row):
        return self.data_by_rows[row][0]



import logging
import os

FORMAT = '[%(levelname)-7s] [%(name)-35s] - %(message)s' 
logging.basicConfig(level=logging.INFO, format=FORMAT)

CAMELOT_MEDIA_ROOT = os.path.join(os.path.dirname(__file__), 'media')
ENGINE = lambda:'sqlite:///videostore.sqlite'

#logging.getLogger('sqlalchemy.engine').setLevel(logging.DEBUG)
#logging.getLogger('camelot.view.proxy.collection_proxy').setLevel(logging.DEBUG)

def setup_model():
    import camelot.model
    import example.model
    from elixir import setup_all
    setup_all(create_tables=True)
    from camelot.model.authentication import updateLastLogin
    updateLastLogin()



from PyQt4 import QtGui, QtCore
from PyQt4.QtCore import Qt

from customeditor import CustomEditor
from camelot.view.art import Icon
from camelot.core import constants

class FloatEditor(CustomEditor):
    """Widget for editing a float field, with a calculator"""
      
    def __init__(self,
                 parent,
                 precision=2,
                 minimum=constants.camelot_minfloat,
                 maximum=constants.camelot_maxfloat,
                 editable=True,
                 prefix='',
                 suffix='',
                 calculator=True,
                 **kwargs):
        CustomEditor.__init__(self, parent)
        action = QtGui.QAction(self)
        action.setShortcut(Qt.Key_F3)
        self.setFocusPolicy(Qt.StrongFocus)
        self.spinBox = QtGui.QDoubleSpinBox(parent)
        self.spinBox.setReadOnly(not editable)
        self.spinBox.setEnabled(editable)
        self.spinBox.setDisabled(not editable)
        self.spinBox.setRange(minimum, maximum)
        self.spinBox.setDecimals(precision)
        self.spinBox.setAlignment(Qt.AlignRight|Qt.AlignVCenter)
        self.spinBox.setSingleStep(1.0)
        
        prefix = str(prefix) + ' '
        prefix = prefix.lstrip()
        
        suffix = ' ' + str(suffix)
        suffix = suffix.rstrip()
        
        self.spinBox.setPrefix(prefix)
        self.spinBox.setSuffix(suffix)
        self.spinBox.addAction(action)
        self.calculatorButton = QtGui.QToolButton()
        icon = Icon('tango/16x16/apps/accessories-calculator.png').getQIcon()
        self.calculatorButton.setIcon(icon)
        self.calculatorButton.setAutoRaise(True)
        self.calculatorButton.setFixedHeight(self.get_height())
        
        self.connect(self.calculatorButton,
                     QtCore.SIGNAL('clicked()'),
                     lambda:self.popupCalculator(self.spinBox.value()))
        self.connect(action,
                     QtCore.SIGNAL('triggered(bool)'),
                     lambda:self.popupCalculator(self.spinBox.value()))
        self.connect(self.spinBox,
                     QtCore.SIGNAL('editingFinished()'),
                     lambda:self.editingFinished(self.spinBox.value()))
    
        self.releaseKeyboard()
    
        layout = QtGui.QHBoxLayout()
        layout.setMargin(0)
        layout.setSpacing(0)
        layout.addWidget(self.spinBox)
        if editable and calculator:
            layout.addWidget(self.calculatorButton)
        if not editable:
            self.spinBox.setEnabled(False)
            self.spinBox.setButtonSymbols(QtGui.QAbstractSpinBox.NoButtons)
        self.setFocusProxy(self.spinBox)
        self.setLayout(layout)
    
    def set_value(self, value):
        value = CustomEditor.set_value(self, value)
        if value:
            self.spinBox.setValue(value)
        else:
            self.spinBox.setValue(0.0)
        
    def set_enabled(self, editable=True):
        if self.spinBox.isEnabled() != editable:
            if not editable:
                self.layout().removeWidget(self.calculatorButton)
            else:
                self.layout().addWidget(self.calculatorButton)
            self.spinBox.setEnabled(editable)
        
    def get_value(self):
        self.spinBox.interpretText()
        value = self.spinBox.value()
        return CustomEditor.get_value(self) or value
    
    def popupCalculator(self, value):
        from camelot.view.controls.calculator import Calculator
        calculator = Calculator(self)
        calculator.setValue(value)
        self.connect(calculator,
                     QtCore.SIGNAL('calculationFinished'),
                     self.calculationFinished)
        calculator.exec_()
    
    def calculationFinished(self, value):
        self.spinBox.setValue(float(value))
        self.emit(QtCore.SIGNAL('editingFinished()'))
    
    def editingFinished(self, value):
        self.emit(QtCore.SIGNAL('editingFinished()'))




"""Camelot includes editors for various types of fields.  Each editor at least supports
these features :


* a set_value method to set a python type as the editor's value
* a get_value method to retrieve a python type from the editor
* the ValueLoading state : an editor has as its value ValueLoading upon construction and
the editor's value can be set to ValueLoading if the value that should be displayed is
not yet available in the GUI thread, but is still on it's way from the model to the GUI. 
"""

from customeditor import editingFinished

from booleditor import BoolEditor
from choiceseditor import ChoicesEditor
from codeeditor import CodeEditor
from coloredfloateditor import ColoredFloatEditor
from coloreditor import ColorEditor
from customeditor import CustomEditor
from dateeditor import DateEditor
from datetimeeditor import DateTimeEditor
from embeddedmany2oneeditor import EmbeddedMany2OneEditor
from fileeditor import FileEditor
from floateditor import FloatEditor
from imageeditor import ImageEditor
from integereditor import IntegerEditor
from many2oneeditor import Many2OneEditor
from manytomanyeditor import ManyToManyEditor
from one2manyeditor import One2ManyEditor
from onetomanychoiceseditor import OneToManyChoicesEditor
from richtexteditor import RichTextEditor
from stareditor import StarEditor
from textlineeditor import TextLineEditor
from timeeditor import TimeEditor
from virtualaddresseditor import VirtualAddressEditor
from smileyeditor import SmileyEditor
from textediteditor import TextEditEditor
from wideeditor import WideEditor




from PyQt4 import QtGui, QtCore
from PyQt4.QtCore import Qt

from customdelegate import CustomDelegate, DocumentationMetaclass
from camelot.view.controls import editors
from camelot.view.art import Icon

class ColoredFloatDelegate(CustomDelegate):
    """Custom delegate for float values, representing them in green when they are
  positive and in red when they are negative.
  """
  
    __metaclass__ = DocumentationMetaclass
    
    editor = editors.ColoredFloatEditor
    
    def __init__(self,
                 parent=None,
                 minimum=-1e15,
                 maximum=1e15,
                 precision=2,
                 editable=True,
                 unicode_format=None,
                 **kwargs):
        CustomDelegate.__init__(self,
                                parent=parent,
                                editable=editable,
                                minimum=minimum,
                                maximum=maximum,
                                precision=precision,
                                unicode_format=unicode_format,
                                **kwargs)
        self.minimum = minimum
        self.maximum = maximum
        self.precision = precision
        self.editable = editable
        self.unicode_format = unicode_format
        
    def paint(self, painter, option, index):
        painter.save()
        self.drawBackground(painter, option, index)
        value = index.model().data(index, Qt.EditRole).toDouble()[0]
        color = index.model().data(index, Qt.BackgroundRole)
        background_color = QtGui.QColor(color)
        fontColor = QtGui.QColor()
        
        if( option.state & QtGui.QStyle.State_Selected ):
            painter.fillRect(option.rect, option.palette.highlight())
        else:
            if not self.editable:
                painter.fillRect(option.rect, option.palette.window())
            else:
                painter.fillRect(option.rect, background_color)
          
          
          
          
            
      
          
        if value >= 0:
            if value == 0:
                QtGui.QIcon(Icon('tango/16x16/actions/zero.png').getQPixmap()).paint(painter, option.rect.top(), option.rect.left()+1, option.rect.height(), option.rect.height(), Qt.AlignVCenter)
                #QtGui.QApplication.style().drawItemPixmap(painter, rect, 1, icon)
                fontColor.setRgb(0, 0, 0)
            else:
                QtGui.QIcon(Icon('tango/16x16/actions/go-up.png').getQPixmap()).paint(painter, option.rect.left(), option.rect.top()+1, option.rect.height(), option.rect.height(), Qt.AlignVCenter)
                #QtGui.QApplication.style().drawItemPixmap(painter, rect, 1, icon)
                fontColor.setRgb(0, 255, 0)
        else:
            QtGui.QIcon(Icon('tango/16x16/actions/go-down-red.png').getQPixmap()).paint(painter, option.rect.left(), option.rect.top()+1, option.rect.height(), option.rect.height(), Qt.AlignVCenter)
#      QtGui.QApplication.style().drawItemPixmap(painter, rect, 1, icon)
            fontColor.setRgb(255, 0, 0)
      
        value_str = '%.*f'%(self.precision, value)
        if self.unicode_format != None:
            value_str = self.unicode_format(value)
            
            
            
        
    
    
    
        fontColor = fontColor.darker()
        painter.setPen(fontColor.toRgb())
        rect = QtCore.QRect(option.rect.left()+23,
                            option.rect.top(),
                            option.rect.width()-23,
                            option.rect.height())
        
        painter.drawText(rect.x()+2,
                         rect.y(),
                         rect.width()-4,
                         rect.height(),
                         Qt.AlignVCenter | Qt.AlignRight,
                         value_str)
        
        painter.restore()



#  ============================================================================
#
#  Copyright (C) 2007-2008 Conceptive Engineering bvba. All rights reserved.
#  www.conceptive.be / project-camelot@conceptive.be
#
#  This file is part of the Camelot Library.
#
#  This file may be used under the terms of the GNU General Public
#  License version 2.0 as published by the Free Software Foundation
#  and appearing in the file LICENSE.GPL included in the packaging of
#  this file.  Please review the following information to ensure GNU
#  General Public Licensing requirements will be met:
#  http://www.trolltech.com/products/qt/opensource.html
#
#  If you are unsure which license is appropriate for your use, please
#  review the following information:
#  http://www.trolltech.com/products/qt/licensing.html or contact
#  project-camelot@conceptive.be.
#
#  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
#  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  For use of this library in commercial applications, please contact
#  project-camelot@conceptive.be
#
#  ============================================================================
from camelot.model import metadata
from camelot.model import entities
from elixir.entity import Entity
from elixir.options import using_options
from elixir.fields import Field
from sqlalchemy.types import Date, Unicode, Integer, DateTime, Boolean
from elixir.relationships import ManyToOne, OneToMany
from elixir.properties import ColumnProperty
from sqlalchemy.sql.expression import and_
"""Set of classes to store persons, organizations, relationships and
permissions

These structures are modeled like described in 'The Data Model Resource Book'
by Len Silverston, Chapter 2
"""
from sqlalchemy import sql
from sqlalchemy.orm import aliased

import camelot.types

#from camelot.model import *

__metadata__ = metadata

from camelot.model.synchronization import is_synchronized
from camelot.core.document import documented_entity
from camelot.core.utils import ugettext_lazy as _

from camelot.view.elixir_admin import EntityAdmin
from camelot.view.forms import Form, TabForm, HBoxForm
from camelot.admin.form_action import FormActionFromModelFunction
import datetime
import threading


_current_authentication_ = threading.local()

def end_of_times():
    return datetime.date( year = 2400, month = 12, day = 31 )

from camelot.model.type_and_status import type_3_status

def getCurrentAuthentication():
    """Get the currently logged in person"""
    global _current_authentication_
    if not hasattr( _current_authentication_, 'mechanism' ):
        import getpass
        _current_authentication_.mechanism = UsernameAuthenticationMechanism.getOrCreateAuthentication( unicode( getpass.getuser() ) )
    return _current_authentication_.mechanism

def updateLastLogin():
    """Update the last login of the current person to now"""
    from elixir import session
    authentication = getCurrentAuthentication()
    authentication.last_login = datetime.datetime.now()
    session.flush( [authentication] )

class PartyRelationship( Entity ):
    using_options( tablename = 'party_relationship' )
    from_date = Field( Date(), default = datetime.date.today, required = True, index = True )
    thru_date = Field( Date(), default = end_of_times, required = True, index = True )
    comment = Field( camelot.types.RichText() )
    is_synchronized( 'synchronized', lazy = True )

    class Admin( EntityAdmin ):
        name = 'Relationship'
        list_display = ['from_date', 'thru_date']

class EmployerEmployee( PartyRelationship ):
    """Relation from employer to employee"""
    using_options( tablename = 'party_relationship_empl', inheritance = 'multi' )
    established_from = ManyToOne( 'Organization', required = True, ondelete = 'cascade', onupdate = 'cascade' )
    established_to = ManyToOne( 'Person', required = True, ondelete = 'cascade', onupdate = 'cascade' )

    @ColumnProperty
    def first_name( self ):
        return sql.select( [Person.first_name], Person.party_id == self.established_to_party_id )

    @ColumnProperty
    def last_name( self ):
        return sql.select( [Person.last_name], Person.party_id == self.established_to_party_id )

    @ColumnProperty
    def social_security_number( self ):
        return sql.select( [Person.social_security_number], Person.party_id == self.established_to_party_id )

    def __unicode__( self ):
        return u'%s %s %s' % ( unicode( self.established_to ), _('employed by'),unicode( self.established_from ) )

    class Admin( EntityAdmin ):
        name = 'Employer - Employee'
        list_display = ['established_from', 'established_to', 'from_date', 'thru_date']

    class EmployeeAdmin( EntityAdmin ):
        verbose_name = 'Employee'
        list_display = ['established_to', 'from_date', 'thru_date']
        form_display = ['established_to', 'comment', 'from_date', 'thru_date']
        field_attributes = {'established_to':{'name':_( 'name' )}}

    class EmployerAdmin( EntityAdmin ):
        verbose_name = 'Employer'
        list_display = ['established_from', 'from_date', 'thru_date']
        form_display = ['established_from', 'comment', 'from_date', 'thru_date']
        field_attributes = {'established_from':{'name':_( 'name' )}}

class DirectedDirector( PartyRelationship ):
    """Relation from a directed organization to a director"""
    using_options( tablename = 'party_relationship_dir', inheritance = 'multi' )
    established_from = ManyToOne( 'Organization', required = True, ondelete = 'cascade', onupdate = 'cascade' )
    established_to = ManyToOne( 'Party', required = True, ondelete = 'cascade', onupdate = 'cascade' )
    title = Field( Unicode( 256 ) )
    represented_by = OneToMany( 'RepresentedRepresentor', inverse = 'established_to' )

    class Admin( EntityAdmin ):
        verbose_name = 'Directed - Director'
        list_display = ['established_from', 'established_to', 'from_date', 'thru_date']

    class DirectorAdmin( EntityAdmin ):
        verbose_name = 'Director'
        list_display = ['established_to', 'from_date', 'thru_date']
        fields = ['established_to', 'title', 'from_date', 'thru_date', 'represented_by', 'comment']
        field_attributes = {'established_to':{'name':_( 'name' )}}

    class DirectedAdmin( EntityAdmin ):
        verbose_name = 'Directed organization'
        list_display = ['established_from', 'from_date', 'thru_date']
        fields = ['established_from', 'from_date', 'thru_date', 'represented_by', 'comment']
        field_attributes = {'established_from':{'name':_( 'name' )}}

class RepresentedRepresentor( Entity ):
    """Relation from a representing party to the person representing the party"""
    using_options( tablename = 'party_representor' )
    from_date = Field( Date(), default = datetime.date.today, required = True, index = True )
    thru_date = Field( Date(), default = end_of_times, required = True, index = True )
    comment = Field( camelot.types.RichText() )
    established_from = ManyToOne( 'Person', required = True, ondelete = 'cascade', onupdate = 'cascade' )
    established_to = ManyToOne( 'DirectedDirector', required = True, ondelete = 'cascade', onupdate = 'cascade' )

    class Admin( EntityAdmin ):
        verbose_name = 'Represented by'
        list_display = ['established_from', 'from_date', 'thru_date']
        form_display = ['established_from', 'from_date', 'thru_date', 'comment']
        field_attributes = {'established_from':{'name':_( 'name' )}}

class SupplierCustomer( PartyRelationship ):
    """Relation from supplier to customer"""
    using_options( tablename = 'party_relationship_suppl', inheritance = 'multi' )
    established_from = ManyToOne( 'Party', required = True, ondelete = 'cascade', onupdate = 'cascade' )
    established_to = ManyToOne( 'Party', required = True, ondelete = 'cascade', onupdate = 'cascade' )

    class Admin( EntityAdmin ):
        verbose_name = 'Supplier - Customer'
        list_display = ['established_from', 'established_to', 'from_date', 'thru_date']

    class CustomerAdmin( EntityAdmin ):
        verbose_name = 'Customer'
        list_display = ['established_to', ]
        fields = ['established_to', 'comment', 'from_date', 'thru_date']
        field_attributes = {'established_to':{'name':_( 'name' )}}

    class SupplierAdmin( EntityAdmin ):
        verbose_name = 'Supplier'
        list_display = ['established_from', ]
        fields = ['established_from', 'comment', 'from_date', 'thru_date']
        field_attributes = {'established_from':{'name':_( 'name' )}}

class SharedShareholder( PartyRelationship ):
    """Relation from a shared organization to a shareholder"""
    using_options( tablename = 'party_relationship_shares', inheritance = 'multi' )
    established_from = ManyToOne( 'Organization', required = True, ondelete = 'cascade', onupdate = 'cascade' )
    established_to = ManyToOne( 'Party', required = True, ondelete = 'cascade', onupdate = 'cascade' )
    shares = Field( Integer() )

    class Admin( EntityAdmin ):
        verbose_name = 'Shared - Shareholder'
        list_display = ['established_from', 'established_to', 'from_date', 'thru_date']

    class ShareholderAdmin( EntityAdmin ):
        verbose_name = 'Shareholder'
        list_display = ['established_to', 'shares', 'from_date', 'thru_date']
        fields = ['established_to', 'shares', 'from_date', 'thru_date', 'comment']
        field_attributes = {'established_to':{'name':'shareholder name'}}

    class SharedAdmin( EntityAdmin ):
        verbose_name = 'Shares'
        verbose_name_plural = 'Shares'
        list_display = ['established_from', 'shares', 'from_date', 'thru_date']
        fields = ['established_from', 'shares', 'from_date', 'thru_date', 'comment']
        field_attributes = {'established_from':{'name':_( 'name' )}}

class AddressAdmin( EntityAdmin ):
    """Admin with only the Address information and not the Party information"""
    list_display = ['address', 'comment']
    form_display = ['address', 'comment', 'from_date', 'thru_date']
    
class PartyContactMechanismAdmin( EntityAdmin ):
    form_size = ( 700, 200 )
    verbose_name = 'Contact mechanism'
    list_search = ['party_name', 'mechanism']
    list_display = ['party_name', 'mechanism', 'comment', 'from_date', ]
    form_display = Form( ['contact_mechanism', 'comment', 'from_date', 'thru_date', ] )
    field_attributes = {'party_name':{'minimal_column_width':25, 'editable':False},
                        'mechanism':{'minimal_column_width':25,'editable':False}}
  
class PartyPartyContactMechanismAdmin( PartyContactMechanismAdmin ):
    list_search = ['party_name', 'mechanism']
    list_display = ['contact_mechanism', 'comment', 'from_date', ]
                
class Party( Entity ):
    """Base class for persons and organizations.  Use this base class to refer to either persons or
    organisations in building authentication systems, contact management or CRM"""
    using_options( tablename = 'party' )
    is_synchronized( 'synchronized', lazy = True )
    addresses = OneToMany( 'PartyAddress', lazy = True )
    contact_mechanisms = OneToMany( 'PartyContactMechanism', lazy = True )
    shares = OneToMany( 'SharedShareholder', inverse = 'established_to' )
    directed_organizations = OneToMany( 'DirectedDirector', inverse = 'established_to' )
    status = OneToMany( type_3_status( 'Party', metadata, entities ) )

    @property
    def name( self ):
        return ''

    @ColumnProperty
    def email( self ):
        
        cm = ContactMechanism
        pcm = PartyContactMechanism
        
        return sql.select( [cm.mechanism],
                          whereclause = and_( pcm.table.c.party_id == self.id,
                                           cm.table.c.mechanism.like( ( u'email', u'%' ) ) ),
                          from_obj = [cm.table.join( pcm.table )] ).limit(1)

    @ColumnProperty
    def phone( self ):
        
        cm = ContactMechanism
        pcm = PartyContactMechanism
                
        return sql.select( [cm.mechanism],
                          whereclause = and_( pcm.table.c.party_id == self.id,
                                           cm.table.c.mechanism.like( ( u'phone', u'%' ) ) ),
                          from_obj = [cm.table.join( pcm.table )] ).limit(1)

    @ColumnProperty
    def full_name( self ):
        aliased_organisation = Organization.table.alias( 'organisation_alias' )
        aliased_person = Person.table.alias( 'person_alias' )
        aliased_party = Party.table.alias( 'party_alias' )
        return sql.functions.coalesce( sql.select( [aliased_person.c.first_name + ' ' + aliased_person.c.last_name],
                                                  whereclause = and_( aliased_party.c.id == self.id ),
                                                  from_obj = [aliased_party.join( aliased_person, aliased_person.c.party_id == aliased_party.c.id )] ).limit( 1 ).as_scalar(),
                                      sql.select( [aliased_organisation.c.name],
                                                 whereclause = and_( aliased_party.c.id == self.id ),
                                                 from_obj = [aliased_party.join( aliased_organisation, aliased_organisation.c.party_id == aliased_party.c.id )] ).limit( 1 ).as_scalar() )

    class Admin( EntityAdmin ):
        verbose_name = 'Party'
        verbose_name_plural = 'Parties'
        list_display = ['name', 'email', 'phone'] # don't use full name, since it might be None for new objects
        list_search = ['full_name']
        fields = ['addresses', 'contact_mechanisms', 'shares', 'directed_organizations']
        field_attributes = dict(addresses = {'admin':AddressAdmin},
                                contact_mechanisms = {'admin':PartyPartyContactMechanismAdmin}, 
                                suppliers = {'admin':SupplierCustomer.SupplierAdmin},
                                customers = {'admin':SupplierCustomer.CustomerAdmin},
                                employers = {'admin':EmployerEmployee.EmployerAdmin},
                                employees = {'admin':EmployerEmployee.EmployeeAdmin},
                                directed_organizations = {'admin':DirectedDirector.DirectedAdmin},
                                directors = {'admin':DirectedDirector.DirectorAdmin},
                                shares = {'admin':SharedShareholder.SharedAdmin},
                                shareholders = {'admin':SharedShareholder.ShareholderAdmin},
                                sex = dict( choices = lambda obj:[( u'M', u'Male' ), ( u'F', u'Female' )], ),
                                name = dict( minimal_column_width = 50 ),
                                email = dict( editable = False, minimal_column_width = 20 ),
                                phone = dict( editable = False, minimal_column_width = 20 )
                                )

class Organization( Party ):
    """An organization represents any internal or external organization.  Organizations can include
    businesses and groups of individuals"""
    using_options( tablename = 'organization', inheritance = 'multi' )
    name = Field( Unicode( 50 ), required = True, index = True )
    logo = Field( camelot.types.Image( upload_to = 'organization-logo' ), deferred = True )
    tax_id = Field( Unicode( 20 ) )
    directors = OneToMany( 'DirectedDirector', inverse = 'established_from' )
    employees = OneToMany( 'EmployerEmployee', inverse = 'established_from' )
    suppliers = OneToMany( 'SupplierCustomer', inverse = 'established_to' )
    customers = OneToMany( 'SupplierCustomer', inverse = 'established_from' )
    shareholders = OneToMany( 'SharedShareholder', inverse = 'established_from' )

    def __unicode__( self ):
        return self.name

    @property
    def number_of_shares_issued( self ):
        return sum( ( shareholder.shares for shareholder in self.shareholders ), 0 )

    class Admin( Party.Admin ):
        verbose_name = _( 'organization' )
        verbose_name_plural = _( 'organizations' )
        section = 'relations'
        list_display = ['name', 'tax_id', 'email', 'phone']
        form_display = TabForm( [( 'Basic', Form( ['name', 'tax_id', 'addresses', 'contact_mechanisms'] ) ),
                                ( 'Employment', Form( ['employees'] ) ),
                                ( 'Customers', Form( ['customers'] ) ),
                                ( 'Suppliers', Form( ['suppliers'] ) ),
                                ( 'Corporate', Form( ['directors', 'shareholders', 'shares'] ) ),
                                ( 'Branding', Form( ['logo'] ) ),
                                ( 'Status', Form( ['status'] ) ),
                                ] )

Organization = documented_entity()( Organization )


        
class AuthenticationMechanism( Entity ):
    using_options( tablename = 'authentication_mechanism' )
    last_login = Field( DateTime() )
    is_active = Field( Boolean, default = True, index = True )

    class Admin( EntityAdmin ):
        verbose_name = 'Authentication mechanism'
        list_display = ['last_login', 'is_active']

class UsernameAuthenticationMechanism( AuthenticationMechanism ):
    using_options( tablename = 'authentication_mechanism_username', inheritance = 'multi' )
    username = Field( Unicode( 40 ), required = True, index = True, unique = True )
    password = Field( Unicode( 200 ), required = False, index = False, default = None )

    @classmethod
    def getOrCreateAuthentication( cls, username ):
        authentication = cls.query.filter_by( username = username ).first()
        if not authentication:
            authentication = cls( username = username )
            from elixir import session
            session.flush( [authentication] )
        return authentication

    def __unicode__( self ):
        return self.username

    class Admin( EntityAdmin ):
        verbose_name = 'Authentication mechanism'
        list_display = ['username', 'last_login', 'is_active']

class Person( Party ):
    """Person represents natural persons
    """
    using_options( tablename = 'person', inheritance = 'multi' )
    first_name = Field( Unicode( 40 ), required = True )
    last_name = Field( Unicode( 40 ), required = True )
    middle_name = Field( Unicode( 40 ) )
    personal_title = Field( Unicode( 10 ) )
    suffix = Field( Unicode( 3 ) )
    sex = Field( Unicode( 1 ), default = u'M' )
    birthdate = Field( Date() )
    martial_status = Field( Unicode( 1 ) )
    social_security_number = Field( Unicode( 12 ) )
    passport_number = Field( Unicode( 20 ) )
    passport_expiry_date = Field( Date() )
    is_staff = Field( Boolean, default = False, index = True )
    is_superuser = Field( Boolean, default = False, index = True )
    picture = Field( camelot.types.Image( upload_to = 'person-pictures' ), deferred = True )
    comment = Field( camelot.types.RichText() )
    employers = OneToMany( 'EmployerEmployee', inverse = 'established_to' )

    @property
    def name( self ):
        # we don't use full name in here, because for new objects, full name will be None, since
        # it needs to be fetched from the db first
        return u'%s %s' % ( self.first_name, self.last_name )

    def __unicode__( self ):
        return self.name

    class Admin( Party.Admin ):
        verbose_name = _( 'person' )
        verbose_name_plural = _( 'persons' )
        list_display = ['first_name', 'last_name', 'email', 'phone']
        form_display = TabForm( [( 'Basic', Form( [HBoxForm( [Form( ['first_name', 'last_name', 'sex'] ),
                                                          Form( ['picture', ] ),
                                                         ] ),
                                                         'contact_mechanisms', 'comment', ], scrollbars = True ) ),
                                ( 'Official', Form( ['birthdate', 'social_security_number', 'passport_number', 'passport_expiry_date', 'addresses', ], scrollbars = True ) ),
                                ( 'Work', Form( ['employers', 'directed_organizations', 'shares'], scrollbars = True ) ),
                                ( 'Status', Form( ['status'] ) ),
                                ] )

Person = documented_entity()( Person )

class GeographicBoundary( Entity ):
    using_options( tablename = 'geographic_boundary' )
    code = Field( Unicode( 10 ) )
    name = Field( Unicode( 40 ), required = True )

    @ColumnProperty
    def full_name( self ):
        return self.code + ' ' + self.name
        
    def __unicode__( self ):
        return u'%s %s' % ( self.code, self.name )

class Country( GeographicBoundary ):
    using_options( tablename = 'geographic_boundary_country', inheritance = 'multi' )

    @classmethod
    def getOrCreate( cls, code, name ):
        country = Country.query.filter_by( code = code ).first()
        if not country:
            from elixir import session
            country = Country( code = code, name = name )
            session.flush( [country] )
        return country

    class Admin( EntityAdmin ):
        form_size = ( 700, 150 )
        verbose_name = 'Country'
        verbose_name_plural = 'Countries'
        list_display = ['name', 'code']

class City( GeographicBoundary ):
    using_options( tablename = 'geographic_boundary_city', inheritance = 'multi' )
    country = ManyToOne( 'Country', required = True, ondelete = 'cascade', onupdate = 'cascade' )

    @classmethod
    def getOrCreate( cls, country, code, name ):
        city = City.query.filter_by( code = code, country = country ).first()
        if not city:
            from elixir import session
            city = City( code = code, name = name, country = country )
            session.flush( [city] )
        return city

    class Admin( EntityAdmin ):
        verbose_name = 'City'
        verbose_name_plural = 'Cities'
        form_size = ( 700, 150 )
        list_display = ['code', 'name', 'country']

class Address( Entity ):
    using_options( tablename = 'address' )
    street1 = Field( Unicode( 128 ), required = True )
    street2 = Field( Unicode( 128 ) )
    city = ManyToOne( 'City', required = True, ondelete = 'cascade', onupdate = 'cascade' )
    is_synchronized( 'synchronized', lazy = True )

    @ColumnProperty
    def name( self ):
        return sql.select( [self.street1 + ', ' + GeographicBoundary.full_name],
                           whereclause = (GeographicBoundary.id==self.city_geographicboundary_id)) 
    
    @classmethod
    def getOrCreate( cls, street1, street2, city ):
        address = cls.query.filter_by( street1 = street1, street2 = street2, city = city ).first()
        if not address:
            from elixir import session
            address = cls( street1 = street1, street2 = street2, city = city )
            session.flush( [address] )
        return address

    def __unicode__( self ):
        return u'%s, %s' % ( self.street1 or '', self.city or '' )

    def showMap( self ):
        from PyQt4 import QtGui, QtCore
        QtGui.QDesktopServices.openUrl ( QtCore.QUrl( 'http://www.google.be/maps?f=q&source=s_q&geocode=%s&q=%s+%s' % ( self.city.country.code, self.street1, self.city.name ) ) )

    class Admin( EntityAdmin ):
        verbose_name = 'Address'
        verbose_name_plural = 'Addresses'
        list_display = ['street1', 'street2', 'city']
        form_size = ( 700, 150 )
        field_attributes = {'street1':{'minimal_column_width':30}}
        form_actions = [FormActionFromModelFunction( 'Show on map', lambda address:address.showMap() )]

Address = documented_entity()( Address )

class PartyAddress( Entity ):
    using_options( tablename = 'party_address' )
    party = ManyToOne( 'Party', required = True, ondelete = 'cascade', onupdate = 'cascade' )
    address = ManyToOne( 'Address', required = True, ondelete = 'cascade', onupdate = 'cascade' )
    from_date = Field( Date(), default = datetime.date.today, required = True, index = True )
    thru_date = Field( Date(), default = end_of_times, required = True, index = True )
    comment = Field( Unicode( 256 ) )

    @ColumnProperty
    def party_name( self ):
        return sql.select( [Party.full_name],
                           whereclause = (Party.id==self.party_id))
        
    @ColumnProperty
    def address_name( self ):
        return sql.select( [Address.name],
                           whereclause = (Address.id==self.address_id))        
                          
    def __unicode__( self ):
        return '%s : %s' % ( unicode( self.party ), unicode( self.address ) )

    def showMap( self ):
        if self.address:
            self.address.showMap()

    class Admin( EntityAdmin ):
        verbose_name = 'Address'
        verbose_name_plural = 'Addresses'
        list_search = ['party_name', 'address_name']
        list_display = ['party_name', 'address_name', 'comment']
        form_display = ['party', 'address', 'comment', 'from_date', 'thru_date']
        form_size = ( 700, 200 )
        form_actions = [FormActionFromModelFunction( 'Show on map', lambda address:address.showMap() )]
        field_attributes = dict(party_name=dict(editable=False, name='Party', minimal_column_width=30),
                                address_name=dict(editable=False, name='Address', minimal_column_width=30))
        
class PartyAddressRoleType( Entity ):
    using_options( tablename = 'party_address_role_type' )
    code = Field( Unicode( 10 ) )
    description = Field( Unicode( 40 ) )

    class Admin( EntityAdmin ):
        verbose_name = 'Address role type'
        list_display = ['code', 'description']

class PartyAuthentication( Entity ):
    using_options( tablename = 'party_authentication' )
    address = ManyToOne( 'AuthenticationMechanism', required = True, ondelete = 'cascade', onupdate = 'cascade' )
    from_date = Field( Date(), default = datetime.date.today, required = True, index = True )
    thru_date = Field( Date(), default = end_of_times, required = True, index = True )
    comment = Field( Unicode( 256 ) )

class ContactMechanism( Entity ):
    using_options( tablename = 'contact_mechanism' )
    mechanism = Field( camelot.types.VirtualAddress( 256 ), required = True )
    party_address = ManyToOne( 'PartyAddress', ondelete = 'set null', onupdate = 'cascade' )

    def __unicode__( self ):
        if self.mechanism:
            return u'%s : %s' % ( self.mechanism[0], self.mechanism[1] )

    class Admin( EntityAdmin ):
        form_size = ( 700, 150 )
        verbose_name = 'Contact mechanism'
        list_display = ['mechanism']
        form_display = Form( ['mechanism', 'party_address'] )
        field_attributes = {'mechanism':{'minimal_column_width':25}}

ContactMechanism = documented_entity()( ContactMechanism )

class PartyContactMechanism( Entity ):
    using_options( tablename = 'party_contact_mechanism' )
    party = ManyToOne( 'Party', required = True, ondelete = 'cascade', onupdate = 'cascade' )
    contact_mechanism = ManyToOne( 'ContactMechanism', required = True, ondelete = 'cascade', onupdate = 'cascade' )
    from_date = Field( Date(), default = datetime.date.today, required = True, index = True )
    thru_date = Field( Date(), default = end_of_times, index = True )
    comment = Field( Unicode( 256 ) )

    @ColumnProperty
    def mechanism( self ):
        return sql.select( [ContactMechanism.mechanism],
                           whereclause = (ContactMechanism.id==self.contact_mechanism_id))
        
    @ColumnProperty
    def party_name( self ):
        return sql.select( [Party.full_name],
                           whereclause = (Party.id==self.party_id))        
        
    def __unicode__( self ):
        return unicode( self.contact_mechanism )

    Admin = PartyContactMechanismAdmin





from customeditor import CustomEditor, QtCore, QtGui
from wideeditor import WideEditor
from camelot.view.model_thread import post
from camelot.view.proxy import ValueLoading

class EmbeddedMany2OneEditor( CustomEditor, WideEditor ):
    """Widget for editing a many 2 one relation a a form embedded in another
  form.
  """
  
    def __init__( self, admin = None, parent = None, **kwargs ):
        assert admin != None
        CustomEditor.__init__( self, parent )
        self.admin = admin
        self.layout = QtGui.QHBoxLayout()
        self.entity_instance_getter = None
        self.form = None
        self.setLayout( self.layout )
        self.setEntity( lambda:ValueLoading, propagate = False )
    
    def set_value( self, value ):
        value = CustomEditor.set_value( self, value )
        if value:
            self.setEntity( value, propagate = False )
      
    def setEntity( self, entity_instance_getter, propagate = True ):
  
        def create_instance_getter( entity_instance ):
            return lambda:entity_instance
      
        def set_entity_instance():
            entity = entity_instance_getter()
            if entity:
                if entity!=ValueLoading:
                    self.entity_instance_getter = create_instance_getter( entity )
                else:
                    return False, False
            else:
                self.entity_instance_getter = create_instance_getter( self.admin.entity() )
            return True, propagate
      
        post( set_entity_instance, self.update_form )
    
    def update_form(self, update_form_and_propagate ):
        update_form, propagate = update_form_and_propagate

        if update_form:
            if self.form:
                self.form.deleteLater()
                self.layout.removeWidget( self.form )
          
            from camelot.view.proxy.collection_proxy import CollectionProxy
        
            def create_collection_getter( instance_getter ):
                return lambda:[instance_getter()]
          
            model = CollectionProxy( self.admin,
                                    create_collection_getter( self.entity_instance_getter ),
                                    self.admin.get_fields )
            self.form = self.admin.create_form_view( '', model, 0, self )
            self.layout.addWidget( self.form )
            
        if propagate:
            self.emit( QtCore.SIGNAL( 'editingFinished()' ) )



#  ============================================================================
#
#  Copyright (C) 2007-2008 Conceptive Engineering bvba. All rights reserved.
#  www.conceptive.be / project-camelot@conceptive.be
#
#  This file is part of the Camelot Library.
#
#  This file may be used under the terms of the GNU General Public
#  License version 2.0 as published by the Free Software Foundation
#  and appearing in the file LICENSE.GPL included in the packaging of
#  this file.  Please review the following information to ensure GNU
#  General Public Licensing requirements will be met:
#  http://www.trolltech.com/products/qt/opensource.html
#
#  If you are unsure which license is appropriate for your use, please
#  review the following information:
#  http://www.trolltech.com/products/qt/licensing.html or contact
#  project-camelot@conceptive.be.
#
#  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
#  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  For use of this library in commercial applications, please contact
#  project-camelot@conceptive.be
#
#  ============================================================================

"""Functionallity common to TableViews and FormViews
"""

from PyQt4 import QtCore, QtGui

from camelot.view.model_thread import post, gui_function

class AbstractView(QtGui.QWidget):
    """A string used to format the title of the view ::

  title_format = 'Movie rental overview'

  .. attribute:: header_widget

  The widget class to be used as a header in the table view::

    header_widget = None
  """

    title_format = ''
    header_widget = None

    title_changed_signal = QtCore.SIGNAL('titleChanged(const QString&)')

    @gui_function
    def change_title(self, new_title):
        """Will emit the title_changed_signal"""
        import sip
        if not sip.isdeleted(self):
            self.emit(self.title_changed_signal, unicode(new_title))

class TabView(AbstractView):
    """Class to combine multiple views in Tabs and let them behave as one view.  This class can be
  used when defining custom create_table_view methods on an ObjectAdmin class to group multiple
  table views together in one view.
  """

    def __init__(self, parent, views=[], admin=None):
        """
    :param views: a list of the views to combine
    """
        AbstractView.__init__(self, parent)
        layout = QtGui.QVBoxLayout()
        if self.header_widget:
            self.header = self.header_widget(self, admin)
        else:
            self.header = None
        layout.addWidget(self.header)
        self.tab_widget = QtGui.QTabWidget(self)
        layout.addWidget(self.tab_widget)
        self.setLayout(layout)

        def get_views_and_titles():
            return [(view, view.get_title()) for view in views]

        post(get_views_and_titles, self.set_views_and_titles )
        post(lambda:self.title_format, self.change_title )

    def set_views_and_titles(self, views_and_titles):
        for view, title in views_and_titles:
            self.tab_widget.addTab(view, title)




from PyQt4.QtCore import Qt
from PyQt4 import QtGui, QtCore

import camelot.types
from camelot.core.constants import *
from camelot.view.proxy import ValueLoading
from camelot.core.utils import create_constant_function

editingFinished = QtCore.SIGNAL('editingFinished()')
 
class AbstractCustomEditor(object):
    """Helper class to be used to build custom editors.  This class provides
  functionallity to store and retrieve `ValueLoading` as an editor's value.
  """
    
    def __init__(self):
        self._value_loading = True
        
    def set_value(self, value):
        if value==ValueLoading:
            self._value_loading = True
            return None
        else:
            self._value_loading = False
            return value
            
    def get_value(self):
        if self._value_loading:
            return ValueLoading
        return None
      
      
    """
    Get the 'standard' height for a cell
    """
    def get_height(self):
      
        height = [QtGui.QLineEdit().sizeHint().height(),
               QtGui.QDateEdit().sizeHint().height(),
               QtGui.QDateTimeEdit().sizeHint().height(),
               QtGui.QSpinBox().sizeHint().height(),
               QtGui.QDateEdit().sizeHint().height(),
               QtGui.QComboBox().sizeHint().height()]
        
        finalHeight = max(height)
        
        return finalHeight
          
class CustomEditor(QtGui.QWidget, AbstractCustomEditor):
    """Base class for implementing custom editor widgets.  This class provides
  dual state functionality.  Each editor should have the posibility to have as
  its value `ValueLoading` specifying that no value has been set yet.
  """
    def __init__(self, parent):
        QtGui.QWidget.__init__(self, parent)
        AbstractCustomEditor.__init__(self)




from PyQt4 import QtGui, QtCore
from PyQt4.QtCore import Qt

from customdelegate import CustomDelegate, DocumentationMetaclass
from camelot.view.controls import editors

class BoolDelegate(CustomDelegate):
    """Custom delegate for boolean values"""
  
    __metaclass__ = DocumentationMetaclass
    
    editor = editors.BoolEditor
  
    def paint(self, painter, option, index):
        painter.save()
        self.drawBackground(painter, option, index)
        checked = index.model().data(index, Qt.EditRole).toBool()
        
        background_color = QtGui.QColor(index.model().data(index, Qt.BackgroundRole))
        
        check_option = QtGui.QStyleOptionButton()
        
        rect = QtCore.QRect(option.rect.left(),
                            option.rect.top(),
                            option.rect.width(),
                            option.rect.height())
        
        check_option.rect = rect
        check_option.palette = option.palette
        if (option.state & QtGui.QStyle.State_Selected):
            painter.fillRect(option.rect, option.palette.highlight())
        elif not self.editable:
            painter.fillRect(option.rect, option.palette.window())
        else:
            painter.fillRect(option.rect, background_color)
            
        if checked:
            check_option.state = option.state | QtGui.QStyle.State_On
        else:
            check_option.state = option.state | QtGui.QStyle.State_Off
            
            
            
        QtGui.QApplication.style().drawControl(QtGui.QStyle.CE_CheckBox,
                                               check_option,
                                               painter)
        
        
        painter.restore()
    
    



from PyQt4 import QtGui
from PyQt4.QtCore import Qt

from customdelegate import CustomDelegate, DocumentationMetaclass
from camelot.view.controls import editors
from camelot.core import constants
from camelot.core.utils import variant_to_pyobject
from camelot.view.proxy import ValueLoading

class FloatDelegate( CustomDelegate ):
    """Custom delegate for float values"""

    __metaclass__ = DocumentationMetaclass

    editor = editors.FloatEditor

    def __init__( self,
                 minimum = constants.camelot_minfloat,
                 maximum = constants.camelot_maxfloat,
                 precision = 2,
                 editable = True,
                 parent = None,
                 unicode_format = None,
                 prefix = '',
                 suffix = '',
                 **kwargs ):
        """
    :param precision:  The number of digits after the decimal point displayed.  This defaults
    to the precision specified in the definition of the Field.     
    """
        CustomDelegate.__init__( self, parent = parent, editable = editable,
                                suffix = suffix, prefix = prefix,
                                precision = precision, **kwargs )
        self.precision = precision
        self.editable = editable
        self.unicode_format = unicode_format
        self.prefix = prefix
        self.suffix = suffix

    def paint( self, painter, option, index ):
        painter.save()
        self.drawBackground( painter, option, index )
        value = variant_to_pyobject( index.model().data( index, Qt.EditRole ) )

        background_color = QtGui.QColor( index.model().data( index, Qt.BackgroundRole ) )

        if( option.state & QtGui.QStyle.State_Selected ):
            painter.fillRect( option.rect, option.palette.highlight() )
            fontColor = QtGui.QColor()
            if self.editable:
                Color = option.palette.highlightedText().color()
                fontColor.setRgb( Color.red(), Color.green(), Color.blue() )
            else:
                fontColor.setRgb( 130, 130, 130 )
        else:
            if self.editable:
                painter.fillRect( option.rect, background_color )
                fontColor = QtGui.QColor()
                fontColor.setRgb( 0, 0, 0 )
            else:
                painter.fillRect( option.rect, option.palette.window() )
                fontColor = QtGui.QColor()
                fontColor.setRgb( 130, 130, 130 )

        value_str = u''
        if value != None and value != ValueLoading:
            value_str = u'%.*f' % ( self.precision, value )

        value_str = unicode( self.prefix ) + ' ' + unicode( value_str ) + ' ' + unicode( self.suffix )
        value_str = value_str.strip()
        if self.unicode_format != None:
            value_str = self.unicode_format( value )

        painter.setPen( fontColor.toRgb() )
        painter.drawText( option.rect.left() + 3,
                         option.rect.top(),
                         option.rect.width() - 6,
                         option.rect.height(),
                         Qt.AlignVCenter | Qt.AlignRight,
                         str( value_str ) )
        painter.restore()



#  ============================================================================
#
#  Copyright (C) 2007-2008 Conceptive Engineering bvba. All rights reserved.
#  www.conceptive.be / project-camelot@conceptive.be
#
#  This file is part of the Camelot Library.
#
#  This file may be used under the terms of the GNU General Public
#  License version 2.0 as published by the Free Software Foundation
#  and appearing in the file LICENSE.GPL included in the packaging of
#  this file.  Please review the following information to ensure GNU
#  General Public Licensing requirements will be met:
#  http://www.trolltech.com/products/qt/opensource.html
#
#  If you are unsure which license is appropriate for your use, please
#  review the following information:
#  http://www.trolltech.com/products/qt/licensing.html or contact
#  project-camelot@conceptive.be.
#
#  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
#  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  For use of this library in commercial applications, please contact
#  project-camelot@conceptive.be
#
#  ============================================================================

"""Controls related to visualizing object hierarchy"""

import logging

logger = logging.getLogger( 'camelot.view.controls.inheritance' )

from PyQt4 import QtGui
from PyQt4 import QtCore

from camelot.view.controls.modeltree import ModelTree, ModelItem
from camelot.view.model_thread import post

QT_MAJOR_VERSION = float( '.'.join( str( QtCore.QT_VERSION_STR ).split( '.' )[0:2] ) )


class SubclassItem( ModelItem ):
    def __init__( self, parent, admin ):
        ModelItem.__init__( self, parent, [admin.get_verbose_name()] )
        self.admin = admin
    
class SubclassTree( ModelTree ):
    """Widget to select subclasses of a certain entity, where the
    subclasses are represented in a tree
    
    emits subclassClicked when a subclass has been selected
    """
  
    def __init__( self, admin, parent ):
        header_labels = ['Types']
        ModelTree.__init__( self, header_labels, parent )
        self.setSizePolicy( QtGui.QSizePolicy.Minimum, QtGui.QSizePolicy.Expanding )
        #self.setSelectionBehavior(QtGui.QAbstractItemView.SelectItems)
        self.admin = admin
        self.subclasses = []
        post( self.admin.get_subclass_tree, self.setSubclasses )
        self.connect( self,
                      QtCore.SIGNAL( 'clicked(const QModelIndex&)' ),
                      self.emitSubclassClicked )
    
    def setSubclasses( self, subclasses ):
        logger.debug( 'setting subclass tree' )
        self.subclasses = subclasses
        
        def append_subclasses(class_item, subclasses):
            for subclass_admin, subsubclasses in subclasses:
                subclass_item = SubclassItem(class_item, subclass_admin)
                self.modelitems.append(subclass_item)
                append_subclasses(subclass_item, subsubclasses)
                
        if len( subclasses ):
            self.clear_model_items()
            top_level_item = SubclassItem( self, self.admin )
            self.modelitems.append( top_level_item )
            append_subclasses(top_level_item, subclasses)
            top_level_item.setExpanded( True )
            self.setMaximumWidth( self.fontMetrics().width( ' ' )*70 )
        else:
            self.setMaximumWidth( 0 )
      
    def emitSubclassClicked( self, index ):
        logger.debug( 'subclass clicked at position %s' % index.row() )
        item = self.itemFromIndex( index )
        self.emit( QtCore.SIGNAL( 'subclassClicked' ), item.admin )
    
class SubclassDialog( QtGui.QDialog ):
    """A dialog requesting the user to select a subclass"""
  
    def __init__( self, parent, admin ):
        QtGui.QDialog.__init__( self, parent )
        layout = QtGui.QVBoxLayout()
        subclass_tree = SubclassTree( admin, self )
        layout.addWidget( subclass_tree )
        layout.addStretch( 1 )
        self.setLayout( layout )
        self.selected_subclass = None
        self.connect( subclass_tree, QtCore.SIGNAL( 'subclassClicked' ), self.subclassClicked )
    
    def subclassClicked( self, admin ):
        self.selected_subclass = admin
        self.accept()




import datetime

from PyQt4 import QtGui, QtCore

from customeditor import CustomEditor
from camelot.core import constants

class DateTimeEditor(CustomEditor):
    """Widget for editing date and time separated and with popups"""
  
    def __init__(self,
                 parent,
                 editable=True,
                 format=constants.camelot_datetime_format,
                 nullable=True,
                 **kwargs):
        CustomEditor.__init__(self, parent)
        import itertools
        self.nullable = nullable
        dateformat, _timeformat = format.split(' ')
        layout = QtGui.QHBoxLayout()
        self.dateedit = QtGui.QDateEdit(self)
        self.dateedit.setEnabled(editable)
        self.dateedit.setDisplayFormat(dateformat)
        self.dateedit.setCalendarPopup(True)
        layout.addWidget(self.dateedit)
            
        class TimeValidator(QtGui.QValidator):
            def __init__(self, parent):
                QtGui.QValidator.__init__(self, parent)
            def validate(self, input, pos):
                parts = str(input).split(':')
                if len(parts)!=2:
                    return (QtGui.QValidator.Invalid, pos)
                if str(input)=='--:--' and nullable:
                    return (QtGui.QValidator.Acceptable, pos)
                for part in parts:
                    if not part.isdigit():
                        return (QtGui.QValidator.Invalid, pos)
                    if len(part) not in (1,2):
                        return (QtGui.QValidator.Intermediate, pos)
                if not int(parts[0]) in range(0,24):
                    return (QtGui.QValidator.Invalid, pos)
                if not int(parts[1]) in range(0,60):
                    return (QtGui.QValidator.Invalid, pos)
                return (QtGui.QValidator.Acceptable, pos)
        
        self.timeedit = QtGui.QComboBox(self)
        self.timeedit.setEditable(True)
        if not editable:
            self.timeedit.setEnabled(False)
        
        time_entries = [entry
                        for entry in itertools.chain(*(('%02i:00'%i, '%02i:30'%i)
                        for i in range(0,24)))]
        self.timeedit.addItems(time_entries)
        self.timeedit.setValidator(TimeValidator(self))
    
#    self.timeedit = QtGui.QTimeEdit(self)
#    self.timeedit.setDisplayFormat(timeformat)
#
#    setting the tab order does not seem to work inside a table
#
#    self.dateedit.setTabOrder(self.dateedit, self.timeedit)

#    Completion doesn't seems to work with a QTimeEdit widget
#
#    time_lineedit = self.timeedit.lineEdit()
#    time_completions_model = QtGui.QStringListModel(['00:00', '00:30'], parent)
#    time_completer = QtGui.QCompleter()
#    time_completer.setModel(time_completions_model)
#    time_completer.setCaseSensitivity(Qt.CaseInsensitive)
#    time_completer.setCompletionMode(QtGui.QCompleter.UnfilteredPopupCompletion)
#    time_lineedit.setCompleter(time_completer)

        layout.addWidget(self.timeedit)
        self.setFocusProxy(self.dateedit)
        self.setLayout(layout)
        layout.setMargin(0)
        layout.setSpacing(0)
        layout.addStretch(1)
            
    def get_value(self):
        time_value = self.time()
        date_value = self.date()
        if time_value!=None and date_value!=None:
            value = datetime.datetime(hour=time_value.hour(),
                                      minute=time_value.minute(),
                                      second=time_value.second(),
                                      year=date_value.year(),
                                      month=date_value.month(),
                                      day=date_value.day())
        else:
            value = None
        return CustomEditor.get_value(self) or value
    
    def set_value(self, value):
        value = CustomEditor.set_value(self, value)
        if value:
            self.dateedit.setDate(QtCore.QDate(value.year, value.month, value.day))
            self.timeedit.lineEdit().setText('%02i:%02i'%(value.hour, value.minute))
        else:
            self.dateedit.setDate(self.dateedit.minimumDate())
            self.timeedit.lineEdit().setText('--:--')
      
    def date(self):
        return self.dateedit.date()
    
    def time(self):
        text = str(self.timeedit.currentText())
        if text=='--:--':
            return None
        parts = text.split(':')
        return QtCore.QTime(int(parts[0]), int(parts[1]))
      
    def set_enabled(self, editable=True):
        self.timeedit.setEnabled(editable)
        self.dateedit.setEnabled(editable)



#  ============================================================================
#
#  Copyright (C) 2007-2008 Conceptive Engineering bvba. All rights reserved.
#  www.conceptive.be / project-camelot@conceptive.be
#
#  This file is part of the Camelot Library.
#
#  This file may be used under the terms of the GNU General Public
#  License version 2.0 as published by the Free Software Foundation
#  and appearing in the file LICENSE.GPL included in the packaging of
#  this file.  Please review the following information to ensure GNU
#  General Public Licensing requirements will be met:
#  http://www.trolltech.com/products/qt/opensource.html
#
#  If you are unsure which license is appropriate for your use, please
#  review the following information:
#  http://www.trolltech.com/products/qt/licensing.html or contact
#  project-camelot@conceptive.be.
#
#  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
#  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  For use of this library in commercial applications, please contact
#  project-camelot@conceptive.be
#
#  ============================================================================

"""left navigation pane"""

import logging
logger = logging.getLogger( 'camelot.view.controls.navpane' )

from PyQt4.QtCore import Qt
from PyQt4 import QtGui, QtCore

from camelot.view.model_thread import post
from camelot.action import addActions, createAction
from camelot.view.controls.modeltree import ModelItem, ModelTree
from appscheme import scheme

QT_MAJOR_VERSION = float( '.'.join( str( QtCore.QT_VERSION_STR ).split( '.' )[0:2] ) )

_ = lambda x:x


class PaneCaption( QtGui.QLabel ):
    """Navigation pane Caption"""
    def __init__( self,
                 text,
                 textbold = True,
                 textindent = 3,
                 width = 160,
                 height = 32,
                 objectname = 'PaneCaption',
                 parent = None ):

        QtGui.QLabel.__init__( self, parent )

        self.setText( text )

        if textbold:
            self.textbold()

        font = self.font()
        font.setPointSize( font.pointSize() + 2 )
        self.setFont( font )

        self.setIndent( textindent )

        self.setObjectName( objectname )

        style = """
        QLabel#PaneCaption {
          margin: 3px 0 0 3px;
          border: 1px solid %s;
          color: %s;
          background-color: %s;
        }
        """ % ( scheme.bordercolor(),
               scheme.captiontextcolor(),
               scheme.captionbackground() )

        self.setStyleSheet( style );
        self.setFixedHeight( height )
        self.resize( width, height )

    def textbold( self ):
        font = self.font()
        font.setBold( True )
        font.setPointSize( font.pointSize() + 1 )
        self.setFont( font )


class PaneButton( QtGui.QWidget ):
    """Custom made navigation pane pushbutton"""
    INDEX = 0  # Keep track of the buttons

    def __init__( self,
                 text,
                 buttonicon = '',
                 textbold = True,
                 textleft = True,
                 width = 160,
                 height = 32,
                 objectname = 'PaneButton',
                 parent = None ):

        QtGui.QWidget.__init__( self, parent )

        if textleft:
            option = QtGui.QBoxLayout.LeftToRight
        else:
            option = QtGui.QBoxLayout.RightToLeft
        self.layout = QtGui.QBoxLayout( option )
        self.layout.setSpacing( 0 )
        self.layout.setContentsMargins( 3, 0, 0, 0 )

        if buttonicon:
            self.icon = QtGui.QLabel()
            self.icon.setPixmap( QtGui.QPixmap( buttonicon ) )
            self.layout.addWidget( self.icon )

        self.label = QtGui.QLabel( text )

        self.layout.addWidget( self.label, 2 )

        self.setLayout( self.layout )

        if textbold:
            self.textbold()

        self.setObjectName( objectname )

        self.stylenormal = """
        QWidget#PaneButton * {
          margin: 0;
          padding-left: 3px;
          color : %s;
          border-color : %s;
          background-color : %s;
        }
        """ % ( scheme.textcolor(),
               scheme.bordercolor(),
               scheme.normalbackground() )

        self.stylehovered = """
        QWidget#PaneButton * {
          margin: 0;
          padding-left: 3px;
          color : %s;
          background-color : %s;
        }
        """ % ( scheme.textcolor(),
               scheme.hoveredbackground() )

        self.styleselected = """
        QWidget#PaneButton * {
          margin: 0;
          padding-left: 3px;
          color : %s;
          background-color : %s;
        }
        """ % ( scheme.selectedcolor(),
               scheme.selectedbackground() )

        self.styleselectedover = """
        QWidget#PaneButton * {
          margin: 0;
          padding-left: 3px;
          color : %s;
          background-color : %s;
        }
        """ % ( scheme.selectedcolor(),
               scheme.selectedbackground( inverted = True ) )

        self.setStyleSheet( self.stylenormal )
        self.setFixedHeight( height )
        self.resize( width, height )
        self.selected = False
        self.index = PaneButton.INDEX
        PaneButton.INDEX += 1

    def textbold( self ):
        font = self.label.font()
        font.setBold( True )
        self.label.setFont( font )

    def enterEvent( self, event ):
        if self.selected:
            self.setStyleSheet( self.styleselectedover )
        else:
            self.setStyleSheet( self.stylehovered )

    def leaveEvent( self, event ):
        if self.selected:
            self.setStyleSheet( self.styleselected )
        else:
            self.setStyleSheet( self.stylenormal )

    def mousePressEvent( self, event ):
        if self.selected:
            return
        else:
            self.selected = True
            self.setStyleSheet( self.styleselectedover )
            # Python shortcut SIGNAL, any object can be passed
            self.emit( QtCore.SIGNAL( 'indexselected' ),
                      ( self.index, self.label.text() ) )


class NavigationPane( QtGui.QDockWidget ):
    """ms office-like navigation pane in Qt"""

    def __init__( self, app_admin, objectname = 'NavigationPane', parent = None ):
        QtGui.QDockWidget.__init__( self, parent )
        self.app_admin = app_admin
        self.setFeatures( QtGui.QDockWidget.NoDockWidgetFeatures )
        self.parent = parent
        self.currentbutton = -1
        self.caption = PaneCaption( '' )
        self.setTitleBarWidget( self.caption )
        self.setObjectName( objectname )
        self.content = QtGui.QWidget()
        self.content.setObjectName( 'NavPaneContent' )
        header_labels = ['']
        layout = QtGui.QVBoxLayout()
        layout.setSpacing( 1 )
        layout.setContentsMargins( 1, 1, 1, 1 )
        self.treewidget = ModelTree( header_labels, self )
        layout.addWidget( self.treewidget )
        self.setMinimumWidth(QtGui.QFontMetrics(QtGui.QApplication.font()).averageCharWidth()*40)
        self.content.setLayout( layout )
        self.setWidget( self.content )
        post(app_admin.get_sections, self.set_sections)
        # Tried selecting QDockWidget but it's not working
        # so we must undo this margin in children stylesheets :)
        #style = 'margin: 0 0 0 3px;'
        #self.setStyleSheet(style)

    def set_sections(self, sections):
        from PyQt4.QtTest import QTest
        self.sections = sections
        self.buttons = [PaneButton( section.get_verbose_name(), section.get_icon().getQPixmap() ) for section in sections]
        self.setcontent( self.buttons )
        # use QTest to auto select first button :)
        if len( self.buttons ):
            firstbutton = self.buttons[0]
            QTest.mousePress( firstbutton, Qt.LeftButton )

    def setcontent( self, buttons ):
        logger.debug( 'setting up pane content' )

        style = """
        QWidget#NavPaneContent {
          margin-left: 3px;
          background-color: %s;
        }
        """ % scheme.bordercolor()

        self.content.setStyleSheet( style )

        # TODO: Should a separator be added between the tree
        #       and the buttons?
        #self.treewidget = PaneTree(self)


        self.treewidget.setObjectName( 'NavPaneTree' )

        style = """
        QWidget#NavPaneTree {
          margin-left: 3px;
          border: 1px solid %s;
          background-color: rgb(255, 255, 255);
        }
        """ % scheme.bordercolor()

        self.treewidget.setStyleSheet( style )

        # context menu
        self.treewidget.contextmenu = QtGui.QMenu( self )
        newWindowAct = createAction( parent = self,
                                    text = _( 'Open in New Window' ),
                                    slot = self.popWindow,
                                    shortcut = 'Ctrl+Enter' )

        addActions( self.treewidget.contextmenu, ( newWindowAct, ) )
        self.treewidget.setContextMenuPolicy( Qt.CustomContextMenu )
        self.connect( self.treewidget,
                     QtCore.SIGNAL( 'customContextMenuRequested(const QPoint &)' ),
                     self.createContextMenu )

        if buttons:
            for b in buttons:
                self.content.layout().addWidget( b )
                self.connect( b, QtCore.SIGNAL( 'indexselected' ), self.change_current )
            self.buttons = buttons
        else:
            self.buttons = []



    def set_items_in_tree( self, items ):
        self.treewidget.clear()
        self.treewidget.clear_model_items()
        self.items = items

        if not items:
            return

        for item in items:
            model_item = ModelItem( self.treewidget, [item.get_verbose_name()] )
            self.treewidget.modelitems.append( model_item )

        self.treewidget.update()

    def change_current( self, ( index, text ) ):
        logger.debug( 'set current to %s' % text )

        if self.currentbutton != -1:
            button = self.buttons[self.currentbutton]
            button.setStyleSheet( button.stylenormal )
            button.selected = False
        self.caption.setText( text )
        self.currentbutton = index

        def get_models_for_tree():
            """Return pairs of (Admin, query) classes for items in the tree"""
            if index<len(self.sections):
                section = self.sections[index]
                return section.get_items()
            return []

        post( get_models_for_tree, self.set_items_in_tree )

    def createContextMenu( self, point ):
        logger.debug( 'creating context menu' )
        item = self.treewidget.itemAt( point )
        if item:
            self.treewidget.setCurrentItem( item )
            self.treewidget.contextmenu.popup( self.cursor().pos() )

    def popWindow( self ):
        """pops a model window in parent's workspace"""
        logger.debug( 'poping a window in parent' )
        item = self.treewidget.currentItem()
        self.parent.createMdiChild( item )



import logging
logger = logging.getLogger( 'camelot.view.export.printer' )

from camelot.view.model_thread import gui_function

@gui_function
def open_html_in_print_preview_from_gui_thread( html ):
    from PyQt4 import QtGui, QtCore
    printer = QtGui.QPrinter()
    printer.setPageSize( QtGui.QPrinter.A4 )
# TODO: make landscape optional
#  printer.setOrientation(QtGui.QPrinter.Landscape)
    # TODO: maximize button
    dialog = QtGui.QPrintPreviewDialog( printer )
  
    def render():
        doc = QtGui.QTextDocument()
        doc.setHtml( html )
        doc.print_( printer )
    
    dialog.connect( dialog, QtCore.SIGNAL( 'paintRequested(QPrinter*)' ), render )
    dialog.exec_()



#  =============================================================================
#
#  Copyright (C) 2007-2008 Conceptive Engineering bvba. All rights reserved.
#  www.conceptive.be / project-camelot@conceptive.be
#
#  This file is part of the Camelot Library.
#
#  This file may be used under the terms of the GNU General Public
#  License version 2.0 as published by the Free Software Foundation
#  and appearing in the file LICENSE.GPL included in the packaging of
#  this file.  Please review the following information to ensure GNU
#  General Public Licensing requirements will be met:
#  http://www.trolltech.com/products/qt/opensource.html
#
#  If you are unsure which license is appropriate for your use, please
#  review the following information:
#  http://www.trolltech.com/products/qt/licensing.html or contact
#  project-camelot@conceptive.be.
#
#  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
#  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  For use of this library in commercial applications, please contact
#  project-camelot@conceptive.be
#
#  =============================================================================
from camelot.model import metadata
import camelot.types
from elixir.entity import Entity
from elixir.options import using_options
from elixir.fields import Field
from sqlalchemy.types import Unicode, INT
"""Set of classes to enable internationalization of the user interface"""


__metadata__ = metadata

from camelot.view.elixir_admin import EntityAdmin
from camelot.core.utils import ugettext_lazy as _

import logging
logger = logging.getLogger( 'camelot.model.i18n' )

def tr( source ):
    from PyQt4 import QtCore
    language = unicode( QtCore.QLocale().name() )
    return Translation.translate_or_register( source, language )

class Translation( Entity ):
    using_options( tablename = 'translation' )
    language = Field( camelot.types.Language, index = True )
    source = Field( Unicode( 500 ), index = True )
    value = Field( Unicode( 500 ) )
    cid = Field( INT(), default = 0, index = True )
    uid = Field( INT(), default = 0, index = True )

    # cache, to prevent too much of the same sql queries
    _cache = dict()

    class Admin( EntityAdmin ):
        verbose_name_plural = _( 'translations' )
        form_size = ( 700, 150 )
        section = 'configuration'
        list_display = ['source', 'language', 'value', 'uid']
        list_filter = ['language']

    @classmethod
    def translate( cls, source, language ):
        """Translate source to language, return None if no translation is found"""
        if source:
            key = ( source, language )
            if key in cls._cache:
                return cls._cache[key]
            translation = cls.query.filter_by( source = unicode( source ), language = language ).filter( Translation.uid != 0 ).first()
            if translation:
                cls._cache[key] = translation.value
                return translation.value
            return None
        return ''

    @classmethod
    def translate_or_register( cls, source, language ):
        """Translate source to language, if no translation is found, register the
        source as to be translated and return the source"""
        if source:
            source = unicode( source )
            translation = cls.translate( source, language )
            if not translation:
                if not cls.query.filter_by( source = source, language = language ).first():
                    if ( source, language ) not in cls._cache:
                        from elixir import session
                        registered_translation = Translation( source = source, language = language )
                        cls._cache[( source, language )] = source
                        session.flush( [registered_translation] )
                        logger.debug( 'registed %s with id %s' % ( source, registered_translation.id ) )
                return source
            return translation
        return ''



#  ============================================================================
#
#  Copyright (C) 2007-2008 Conceptive Engineering bvba. All rights reserved.
#  www.conceptive.be / project-camelot@conceptive.be
#
#  This file is part of the Camelot Library.
#
#  This file may be used under the terms of the GNU General Public
#  License version 2.0 as published by the Free Software Foundation
#  and appearing in the file LICENSE.GPL included in the packaging of
#  this file.  Please review the following information to ensure GNU
#  General Public Licensing requirements will be met:
#  http://www.trolltech.com/products/qt/opensource.html
#
#  If you are unsure which license is appropriate for your use, please
#  review the following information:
#  http://www.trolltech.com/products/qt/licensing.html or contact
#  project-camelot@conceptive.be.
#
#  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
#  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  For use of this library in commercial applications, please contact
#  project-camelot@conceptive.be
#
#  ============================================================================

"""Proxies representing the results of a query"""

from PyQt4.QtCore import Qt

import logging
logger = logging.getLogger('camelot.view.proxy.queryproxy')

from collection_proxy import CollectionProxy, stripped_data_to_unicode, \
                             strip_data_from_object, tool_tips_from_object, \
                             background_colors_from_object
from camelot.view.model_thread import model_function, gui_function


class QueryTableProxy(CollectionProxy):
    """The QueryTableProxy contains a limited copy of the data in the Elixir
    model, which is fetched from the database to be used as the model for a
    QTableView
    """

    def __init__(self, admin, query_getter, columns_getter,
                 max_number_of_rows=10, edits=None):
        """@param query_getter: a model_thread function that returns a query"""
        logger.debug('initialize query table')
        self._query_getter = query_getter
        #rows appended to the table which have not yet been flushed to the
        #database, and as such cannot be a result of the query
        self._appended_rows = []
        CollectionProxy.__init__(self, admin, lambda: [],
                                 columns_getter, max_number_of_rows=10, edits=None)

    @model_function
    def _clean_appended_rows(self):
        """Remove those rows from appended rows that have been flushed"""
        flushed_rows = []
        for o in self._appended_rows:
            if o.id:
                flushed_rows.append(o)
        for o in flushed_rows:
            self._appended_rows.remove(o)

    @model_function
    def getRowCount(self):
        self._clean_appended_rows()
        return self._query_getter().count() + len(self._appended_rows)

    @gui_function
    def setQuery(self, query_getter):
        """Set the query and refresh the view"""
        self._query_getter = query_getter
        self.refresh()

    def append(self, o):
        """Add an object to this collection, used when inserting a new
        row, overwrite this method for specific behaviour in subclasses"""
        if not o.id:
            self._appended_rows.append(o)
        self.rows = self.rows + 1

    def remove(self, o):
        if o in self._appended_rows:
            self._appended_rows.remove(o)
        self.rows = self.rows - 1

    @model_function
    def getData(self):
        """Generator for all the data queried by this proxy"""
        for _i,o in enumerate(self._query_getter().all()):
            yield strip_data_from_object(o, self.getColumns())

    @model_function
    def _extend_cache(self, offset, limit):
        """Extend the cache around row"""
        q = self._query_getter().offset(offset).limit(limit)
        columns = self.getColumns()
        for i, o in enumerate(q.all()):
            row_data = strip_data_from_object(o, columns)
            self.cache[Qt.EditRole].add_data(i+offset, o, row_data)
            self.cache[Qt.ToolTipRole].add_data(i+offset, o, tool_tips_from_object(o, columns))
            self.cache[Qt.BackgroundColorRole].add_data(i+offset, o, background_colors_from_object(o, columns))
            self.cache[Qt.DisplayRole].add_data(i+offset, o, stripped_data_to_unicode(row_data, o, columns))
        rows_in_query = (self.rows - len(self._appended_rows))
        # Verify if rows that have not yet been flushed have been requested
        if offset+limit>=rows_in_query:
            for row in range(max(rows_in_query,offset), min(offset+limit, self.rows)):
                o = self._get_object(row)
                row_data = strip_data_from_object(o, columns)
                self.cache[Qt.EditRole].add_data(row, o, row_data)
                self.cache[Qt.ToolTipRole].add_data(row, o, tool_tips_from_object(o, columns))
                self.cache[Qt.BackgroundColorRole].add_data(row, o, background_colors_from_object(o, columns))
                self.cache[Qt.DisplayRole].add_data(row, o, stripped_data_to_unicode(row_data, o, columns))
        return (offset, limit)

    @model_function
    def _get_object(self, row):
        """Get the object corresponding to row"""
        if self.rows > 0:
            self._clean_appended_rows()
            rows_in_query = (self.rows - len(self._appended_rows))
            if row >= rows_in_query:
                return self._appended_rows[row - rows_in_query]
            # first try to get the primary key out of the cache, if it's not
            # there, query the collection_getter
            try:
                return self.cache[Qt.EditRole].get_entity_at_row(row)
            except KeyError:
                pass
            # momentary hack for list error that prevents forms to be closed
            res = self._query_getter().offset(row)
            if isinstance(res, list):
                res = res[0]
            # @todo: remove this try catch and find out why it sometimes fails
            try:
                return res.limit(1).first()
            except:
                pass



from customdelegate import CustomDelegate
from camelot.view.controls import editors

class ManyToOneChoicesDelegate( CustomDelegate ):
    """Display a ManyToOne field as a ComboBox, filling the list of choices with
  the objects of the target class. 
  
  .. image:: ../_static/enumeration.png   
  """
  
    editor = editors.OneToManyChoicesEditor



from customeditor import CustomEditor, QtCore, QtGui, Qt
from wideeditor import WideEditor
from camelot.view.art import Icon

class RichTextEditor(CustomEditor, WideEditor):
  
    def __init__(self, parent=None, editable=True, **kwargs):
        CustomEditor.__init__(self, parent)
        self.layout = QtGui.QVBoxLayout(self)
        self.layout.setSpacing(0)
        self.layout.setMargin(0)
        self.editable = editable
    
        class CustomTextEdit(QtGui.QTextEdit):
            """A TextEdit editor that sends editingFinished events when the text was changed
            and focus is lost
            """
            
            def __init__(self, parent):
                super(CustomTextEdit, self).__init__(parent)
                self._changed = False
                self.connect(self, QtCore.SIGNAL('textChanged()'), self.setTextChanged)
        
            def focusOutEvent(self, event):
                if self._changed:
                    self.emit(QtCore.SIGNAL('editingFinished()'))
              
            def textChanged(self):
                return self._changed
              
            def setTextChanged(self, state=True):
                self._changed = state
                
            def setHtml(self, html):
                QtGui.QTextEdit.setHtml(self, html)
                self._changed = False
                
        self.textedit = CustomTextEdit(self)
        
        self.connect(self.textedit,
                     QtCore.SIGNAL('editingFinished()'),
                     self.editingFinished)
        self.textedit.setAcceptRichText(True)
    
        if not self.editable:
            self.textedit.setReadOnly(True)
        else:
            self.initButtons(editable)
            
            
#      #
#      # Buttons setup
#      #
#      self.toolbar = QtGui.QToolBar(self)
#      self.toolbar.setContentsMargins(0, 0, 0, 0)
#      self.toolbar.setEnabled(editable)
#      
#      self.bold_button = QtGui.QToolButton(self)
#      icon = Icon('tango/16x16/actions/format-text-bold.png').getQIcon()
#      self.bold_button.setIcon(icon)
#      self.bold_button.setAutoRaise(True)
#      self.bold_button.setCheckable(True)
#      self.bold_button.setEnabled(editable)
#      self.bold_button.setMaximumSize(QtCore.QSize(20, 20))
#      self.bold_button.setShortcut(QtGui.QKeySequence('Ctrl+B'))
#      self.connect(self.bold_button, QtCore.SIGNAL('clicked()'), self.set_bold)
#
#      self.italic_button = QtGui.QToolButton(self)
#      icon = Icon('tango/16x16/actions/format-text-italic.png').getQIcon()
#      self.italic_button.setIcon(icon)
#      self.italic_button.setAutoRaise(True)
#      self.italic_button.setCheckable(True)
#      self.italic_button.setEnabled(editable)
#      self.italic_button.setMaximumSize(QtCore.QSize(20, 20))
#      self.italic_button.setShortcut(QtGui.QKeySequence('Ctrl+I'))
#      self.connect(self.italic_button,
#                   QtCore.SIGNAL('clicked(bool)'),
#                   self.set_italic)
#  
#      self.underline_button = QtGui.QToolButton(self)
#      icon = Icon('tango/16x16/actions/format-text-underline.png').getQIcon()
#      self.underline_button.setIcon(icon)
#      self.underline_button.setAutoRaise(True)
#      self.underline_button.setCheckable(True)
#      self.underline_button.setEnabled(editable)
#      self.underline_button.setMaximumSize(QtCore.QSize(20, 20))
#      self.underline_button.setShortcut(QtGui.QKeySequence('Ctrl+U'))
#      self.connect(self.underline_button,
#                   QtCore.SIGNAL('clicked(bool)'),
#                   self.set_underline)
#  
#      self.copy_button = QtGui.QToolButton(self)
#      icon = Icon('tango/16x16/actions/edit-copy.png').getQIcon()
#      self.copy_button.setIcon(icon)
#      self.copy_button.setAutoRaise(True)
#      self.copy_button.setEnabled(editable)
#      self.copy_button.setMaximumSize(QtCore.QSize(20, 20))
#      self.connect(self.copy_button,
#                   QtCore.SIGNAL('clicked(bool)'),
#                   self.textedit.copy)
#  
#      self.cut_button = QtGui.QToolButton(self)
#      icon = Icon('tango/16x16/actions/edit-cut.png').getQIcon()
#      self.cut_button.setIcon(icon)
#      self.cut_button.setAutoRaise(True)
#      self.cut_button.setEnabled(editable)
#      self.cut_button.setMaximumSize(QtCore.QSize(20, 20))
#      self.connect(self.cut_button,
#                   QtCore.SIGNAL('clicked(bool)'),
#                   self.textedit.cut)
#  
#      self.paste_button = QtGui.QToolButton(self)
#      icon = Icon('tango/16x16/actions/edit-paste.png').getQIcon()
#      self.paste_button.setIcon(icon)
#      self.paste_button.setAutoRaise(True)
#      self.paste_button.setEnabled(editable)
#      self.paste_button.setMaximumSize(QtCore.QSize(20, 20))
#      self.connect(self.paste_button,
#                   QtCore.SIGNAL('clicked(bool)'),
#                   self.textedit.paste)
#  
#      self.alignleft_button = QtGui.QToolButton(self)
#      icon = Icon('tango/16x16/actions/format-justify-left.png').getQIcon()
#      self.alignleft_button.setIcon(icon)
#      self.alignleft_button.setAutoRaise(True)
#      self.alignleft_button.setCheckable(True)
#      self.alignleft_button.setEnabled(editable)
#      self.alignleft_button.setMaximumSize(QtCore.QSize(20, 20))
#      self.connect(self.alignleft_button,
#                   QtCore.SIGNAL('clicked(bool)'),
#                   self.set_alignleft)   
#  
#      self.aligncenter_button = QtGui.QToolButton(self)
#      icon = Icon('tango/16x16/actions/format-justify-center.png').getQIcon()
#      self.aligncenter_button.setIcon(icon)
#      self.aligncenter_button.setAutoRaise(True)
#      self.aligncenter_button.setCheckable(True)
#      self.aligncenter_button.setEnabled(editable)
#      self.aligncenter_button.setMaximumSize(QtCore.QSize(20, 20))
#      self.connect(self.aligncenter_button,
#                   QtCore.SIGNAL('clicked(bool)'),
#                   self.set_aligncenter)
#  
#      self.alignright_button = QtGui.QToolButton(self)
#      icon = Icon('tango/16x16/actions/format-justify-right.png').getQIcon()
#      self.alignright_button.setIcon(icon)
#      self.alignright_button.setAutoRaise(True)
#      self.alignright_button.setCheckable(True)
#      self.alignright_button.setEnabled(editable)
#      self.alignright_button.setMaximumSize(QtCore.QSize(20, 20))
#      self.connect(self.alignright_button,
#                   QtCore.SIGNAL('clicked(bool)'),
#                   self.set_alignright)
#  
#      self.color_button = QtGui.QToolButton(self)
#      self.color_button.setAutoRaise(True)
#      self.color_button.setEnabled(editable)
#      self.color_button.setMaximumSize(QtCore.QSize(20, 20))
#      self.connect(self.color_button,
#                   QtCore.SIGNAL('clicked(bool)'),
#                   self.set_color)
# 
#      self.toolbar.addWidget(self.copy_button)
#      self.toolbar.addWidget(self.cut_button)
#      self.toolbar.addWidget(self.paste_button)
#      self.toolbar.addSeparator()
#      self.toolbar.addWidget(self.bold_button)
#      self.toolbar.addWidget(self.italic_button)      
#      self.toolbar.addWidget(self.underline_button) 
#      self.toolbar.addSeparator()
#      self.toolbar.addWidget(self.alignleft_button)
#      self.toolbar.addWidget(self.aligncenter_button)      
#      self.toolbar.addWidget(self.alignright_button)   
#      self.toolbar.addSeparator()
#      self.toolbar.addWidget(self.color_button)   
#      
#      #
#      # Layout
#      #
#      self.layout.addWidget(self.toolbar)
        self.layout.addWidget(self.textedit)
       
        self.setLayout(self.layout)
        
        #
        # Format
        #
        self.textedit.setFontWeight(QtGui.QFont.Normal)
        self.textedit.setFontItalic(False)
        self.textedit.setFontUnderline(False)
        self.textedit.setFocus(Qt.OtherFocusReason)
        self.textedit.setEnabled(editable)
        self.update_alignment()
    
        if self.editable:
            self.connect(self.textedit,
                         QtCore.SIGNAL('currentCharFormatChanged(const QTextCharFormat&)'),
                         self.update_format)
            self.connect(self.textedit,
                         QtCore.SIGNAL('cursorPositionChanged()'),
                         self.update_text)
                
    def editingFinished(self):
        if self.textedit.textChanged():
            self.emit(QtCore.SIGNAL('editingFinished()'))
            
        
        
        
        
        
    def initButtons(self, editable=True):
        self.toolbar = QtGui.QToolBar(self)
        self.toolbar.setContentsMargins(0, 0, 0, 0)
        self.toolbar.setEnabled(editable)
        
        self.bold_button = QtGui.QToolButton(self)
        icon = Icon('tango/16x16/actions/format-text-bold.png').getQIcon()
        self.bold_button.setIcon(icon)
        self.bold_button.setAutoRaise(True)
        self.bold_button.setCheckable(True)
        self.bold_button.setEnabled(editable)
        self.bold_button.setMaximumSize(QtCore.QSize(20, 20))
        self.bold_button.setShortcut(QtGui.QKeySequence('Ctrl+B'))
        self.connect(self.bold_button, QtCore.SIGNAL('clicked()'), self.set_bold)
    
        self.italic_button = QtGui.QToolButton(self)
        icon = Icon('tango/16x16/actions/format-text-italic.png').getQIcon()
        self.italic_button.setIcon(icon)
        self.italic_button.setAutoRaise(True)
        self.italic_button.setCheckable(True)
        self.italic_button.setEnabled(editable)
        self.italic_button.setMaximumSize(QtCore.QSize(20, 20))
        self.italic_button.setShortcut(QtGui.QKeySequence('Ctrl+I'))
        self.connect(self.italic_button,
                     QtCore.SIGNAL('clicked(bool)'),
                     self.set_italic)
    
        self.underline_button = QtGui.QToolButton(self)
        icon = Icon('tango/16x16/actions/format-text-underline.png').getQIcon()
        self.underline_button.setIcon(icon)
        self.underline_button.setAutoRaise(True)
        self.underline_button.setCheckable(True)
        self.underline_button.setEnabled(editable)
        self.underline_button.setMaximumSize(QtCore.QSize(20, 20))
        self.underline_button.setShortcut(QtGui.QKeySequence('Ctrl+U'))
        self.connect(self.underline_button,
                     QtCore.SIGNAL('clicked(bool)'),
                     self.set_underline)
    
        self.copy_button = QtGui.QToolButton(self)
        icon = Icon('tango/16x16/actions/edit-copy.png').getQIcon()
        self.copy_button.setIcon(icon)
        self.copy_button.setAutoRaise(True)
        self.copy_button.setEnabled(editable)
        self.copy_button.setMaximumSize(QtCore.QSize(20, 20))
        self.connect(self.copy_button,
                     QtCore.SIGNAL('clicked(bool)'),
                     self.textedit.copy)
    
        self.cut_button = QtGui.QToolButton(self)
        icon = Icon('tango/16x16/actions/edit-cut.png').getQIcon()
        self.cut_button.setIcon(icon)
        self.cut_button.setAutoRaise(True)
        self.cut_button.setEnabled(editable)
        self.cut_button.setMaximumSize(QtCore.QSize(20, 20))
        self.connect(self.cut_button,
                     QtCore.SIGNAL('clicked(bool)'),
                     self.textedit.cut)
    
        self.paste_button = QtGui.QToolButton(self)
        icon = Icon('tango/16x16/actions/edit-paste.png').getQIcon()
        self.paste_button.setIcon(icon)
        self.paste_button.setAutoRaise(True)
        self.paste_button.setEnabled(editable)
        self.paste_button.setMaximumSize(QtCore.QSize(20, 20))
        self.connect(self.paste_button,
                     QtCore.SIGNAL('clicked(bool)'),
                     self.textedit.paste)
    
        self.alignleft_button = QtGui.QToolButton(self)
        icon = Icon('tango/16x16/actions/format-justify-left.png').getQIcon()
        self.alignleft_button.setIcon(icon)
        self.alignleft_button.setAutoRaise(True)
        self.alignleft_button.setCheckable(True)
        self.alignleft_button.setEnabled(editable)
        self.alignleft_button.setMaximumSize(QtCore.QSize(20, 20))
        self.connect(self.alignleft_button,
                     QtCore.SIGNAL('clicked(bool)'),
                     self.set_alignleft)   
    
        self.aligncenter_button = QtGui.QToolButton(self)
        icon = Icon('tango/16x16/actions/format-justify-center.png').getQIcon()
        self.aligncenter_button.setIcon(icon)
        self.aligncenter_button.setAutoRaise(True)
        self.aligncenter_button.setCheckable(True)
        self.aligncenter_button.setEnabled(editable)
        self.aligncenter_button.setMaximumSize(QtCore.QSize(20, 20))
        self.connect(self.aligncenter_button,
                     QtCore.SIGNAL('clicked(bool)'),
                     self.set_aligncenter)
    
        self.alignright_button = QtGui.QToolButton(self)
        icon = Icon('tango/16x16/actions/format-justify-right.png').getQIcon()
        self.alignright_button.setIcon(icon)
        self.alignright_button.setAutoRaise(True)
        self.alignright_button.setCheckable(True)
        self.alignright_button.setEnabled(editable)
        self.alignright_button.setMaximumSize(QtCore.QSize(20, 20))
        self.connect(self.alignright_button,
                     QtCore.SIGNAL('clicked(bool)'),
                     self.set_alignright)
    
        self.color_button = QtGui.QToolButton(self)
        self.color_button.setAutoRaise(True)
        self.color_button.setEnabled(editable)
        self.color_button.setMaximumSize(QtCore.QSize(20, 20))
        self.connect(self.color_button,
                     QtCore.SIGNAL('clicked(bool)'),
                     self.set_color)
    
        self.toolbar.addWidget(self.copy_button)
        self.toolbar.addWidget(self.cut_button)
        self.toolbar.addWidget(self.paste_button)
        self.toolbar.addSeparator()
        self.toolbar.addWidget(self.bold_button)
        self.toolbar.addWidget(self.italic_button)      
        self.toolbar.addWidget(self.underline_button) 
        self.toolbar.addSeparator()
        self.toolbar.addWidget(self.alignleft_button)
        self.toolbar.addWidget(self.aligncenter_button)      
        self.toolbar.addWidget(self.alignright_button)   
        self.toolbar.addSeparator()
        self.toolbar.addWidget(self.color_button)   
        
        #
        # Layout
        #
        self.layout.addWidget(self.toolbar)
    
      
      
      
    def set_enabled(self, editable=True):
        if self.textedit.isEnabled() != editable:
            if not editable:
                self.layout().removeWidget(self.bold_button)
                self.layout().removeWidget(self.italic_button)
                self.layout().removeWidget(self.underline_button)
                self.layout().removeWidget(self.toolbar)
                self.layout().removeWidget(self.color_button)
                self.layout().removeWidget(self.aligncenter_button)
                self.layout().removeWidget(self.alignleft_button)
                self.layout().removeWidget(self.alignright_button)
                self.layout().removeWidget(self.copy_button)
                self.layout().removeWidget(self.cut_button)
                self.layout().removeWidget(self.paste_button)
            else:
                value = self.get_value()
                self.layout.removeWidget(self.textedit)
                self.initButtons(editable)
                self.layout.addWidget(self.textedit)
                self.set_value(value)
            self.textedit.setEnabled(editable)
          
        
      
    #
    # Button methods
    #
    def set_bold(self):
        if self.bold_button.isChecked():
            self.textedit.setFocus(Qt.OtherFocusReason)
            self.textedit.setFontWeight(QtGui.QFont.Bold)
        else:
            self.textedit.setFocus(Qt.OtherFocusReason)
            self.textedit.setFontWeight(QtGui.QFont.Normal)
      
    def set_italic(self, bool):
        if bool:
            self.textedit.setFocus(Qt.OtherFocusReason)
            self.textedit.setFontItalic(True)
        else:
            self.textedit.setFocus(Qt.OtherFocusReason)
            self.textedit.setFontItalic(False)
      
    def set_underline(self, bool):
        if bool:
            self.textedit.setFocus(Qt.OtherFocusReason)
            self.textedit.setFontUnderline(True)
        else:
            self.textedit.setFocus(Qt.OtherFocusReason)
            self.textedit.setFontUnderline(False)
      
      
    def set_alignleft(self, bool):
        if bool:
            self.textedit.setFocus(Qt.OtherFocusReason)
            self.textedit.setAlignment(Qt.AlignLeft)
        self.update_alignment(Qt.AlignLeft)
    
    def set_aligncenter(self, bool):
        if bool:
            self.textedit.setFocus(Qt.OtherFocusReason)
            self.textedit.setAlignment(Qt.AlignCenter)
        self.update_alignment(Qt.AlignCenter)
    
    def set_alignright(self, bool):
        if bool:
            self.textedit.setFocus(Qt.OtherFocusReason)
            self.textedit.setAlignment(Qt.AlignRight)
        self.update_alignment(Qt.AlignRight)
    
    def update_alignment(self, al=None):
        if self.editable:
            if al is None:
                al = self.textedit.alignment()
            if al == Qt.AlignLeft:
                self.alignleft_button.setChecked(True)
                self.aligncenter_button.setChecked(False)
                self.alignright_button.setChecked(False)
            elif al == Qt.AlignCenter:
                self.aligncenter_button.setChecked(True)
                self.alignleft_button.setChecked(False)
                self.alignright_button.setChecked(False)
            elif al == Qt.AlignRight:
                self.alignright_button.setChecked(True)
                self.alignleft_button.setChecked(False)
                self.aligncenter_button.setChecked(False)
      
    def set_color(self):
        color = QtGui.QColorDialog.getColor(self.textedit.textColor())
        if color.isValid():
            self.textedit.setFocus(Qt.OtherFocusReason)
            self.textedit.setTextColor(color)
            pixmap = QtGui.QPixmap(16, 16)
            pixmap.fill(color)
            self.color_button.setIcon(QtGui.QIcon(pixmap))
        
    def update_color(self):
        if self.editable:
            color = self.textedit.textColor()
            pixmap = QtGui.QPixmap(16, 16)
            pixmap.fill(color)
            self.color_button.setIcon(QtGui.QIcon(pixmap))
      
    def update_format(self, format):
        if self.editable:
            font = format.font()
            self.bold_button.setChecked(font.bold())
            self.italic_button.setChecked(font.italic())
            self.underline_button.setChecked(font.underline())
            self.update_alignment(self.textedit.alignment())
      
    def update_text(self):
        if self.editable:
            self.update_alignment()
            self.update_color()
      
    def get_value(self):
        from xml.dom import minidom
        tree = minidom.parseString(unicode(self.textedit.toHtml()))
        value = u''.join([node.toxml() for node in tree.getElementsByTagName('html')[0].getElementsByTagName('body')[0].childNodes])
        return CustomEditor.get_value(self) or value
        
    def set_value(self, value):
        value = CustomEditor.set_value(self, value)
        if value!=None:
            if unicode(self.textedit.toHtml())!=value:
                self.update_alignment()
                self.textedit.setHtml(value)
                self.update_color()
        else:
            self.textedit.clear()



from PyQt4 import QtGui, QtCore
from camelot.view.art import Icon
from camelot.view.model_thread import post

class ListAction( object ):
    """Abstract base class to implement list actions"""

    def __init__( self, name, icon = None ):
        self._name = name
        self._icon = icon

    def render( self, parent, collection_getter, selection_getter ):
        """Returns a QWidget the user can use to trigger the action"""

        def create_clicked_function( self, collection_getter, selection_getter ):

            def clicked( *args ):
                self.run( collection_getter, selection_getter )

            return clicked

        button = QtGui.QPushButton( unicode(self._name) )
        if self._icon:
            button.setIcon( self._icon.getQIcon() )
        button.connect( button, QtCore.SIGNAL( 'clicked()' ), create_clicked_function( self, collection_getter, selection_getter ) )
        return button

    def run( self, collection_getter, selection_getter ):
        """Overwrite this method to create an action that does something"""
        raise NotImplementedError

class ListActionFromGuiFunction( ListAction ):
    """Convert a function that is supposed to run in the GUI thread to a ListAction"""

    def __init__( self, name, gui_function, icon = None ):
        ListAction.__init__( self, name, icon )
        self._gui_function = gui_function

    def run( self, collection_getter, selection_getter ):
        self._gui_function( collection_getter, selection_getter )

class ListActionFromModelFunction( ListAction ):
    """Convert a function that is supposed to run in the model thread to a FormAction"""

    def __init__( self, name, model_function, icon = None ):
        ListAction.__init__( self, name, icon )
        self._model_function = model_function

    def run( self, collection_getter, selection_getter ):
        progress = QtGui.QProgressDialog( 'Please wait', QtCore.QString(), 0, 0 )
        progress.setWindowTitle( unicode(self._name) )
        progress.show()

        def create_request( collection_getter ):

            def request():
                c = collection_getter()
                s = selection_getter()
                self._model_function( c, s )

            return request

        post( create_request( collection_getter ), progress.close, exception = progress.close )

class PrintHtmlListAction( ListActionFromModelFunction ):

    def __init__( self, name, icon = Icon( 'tango/22x22/actions/document-print.png' ) ):

        def model_function( collection, selection ):
            from camelot.view.export.printer import open_html_in_print_preview
            html = self.html( collection, selection )
            open_html_in_print_preview( html )

        ListActionFromModelFunction.__init__( self, name, model_function, icon )

    def html( self, collection, selection ):
        """Overwrite this function to generate custom html to be printed
    :arg collection: the collection of objects displayed in the list
    :arg selection: the collection of selected objects in the list
        """
        return '<br/>'.join( list( unicode( o ) for o in collection ) )

def structure_to_list_actions( structure ):
    """Convert a list of python objects to a list of list actions.  If the python
    object is a tuple, a ListActionFromGuiFunction is constructed with this tuple as arguments.  If
    the python object is an instance of a ListAction, it is kept as is.
    """

    def object_to_action( o ):
        if isinstance( o, ListAction ):
            return o
        return ListActionFromGuiFunction( o[0], o[1] )

    return [object_to_action( o ) for o in structure]
